#!/usr/bin/env python3

"""
Runs pylint on the RETURNN source code.
"""

import os
import sys
import subprocess

my_dir = os.path.dirname(os.path.abspath(__file__))
base_dir = os.path.dirname(my_dir)
sys.path.insert(0, base_dir)
os.chdir(base_dir)

from returnn.util import better_exchook  # noqa
from returnn.util.basic import pip_install, pip_check_is_installed  # noqa

py = sys.executable


def setup():
  """
  Some generic setup.
  """
  # travis_fold: https://github.com/travis-ci/travis-ci/issues/1065
  print("travis_fold:start:script.install")
  for pkg in ["pylint", "better-exchook"]:
    if not pip_check_is_installed(pkg):
      pip_install(pkg)
  print("travis_fold:end:script.install")


def main():
  """
  Main entry point.
  """
  setup()
  from lint_common import ignore_count_for_files, find_all_py_source_files
  color = better_exchook.Color()
  num_relevant_files_with_errors = 0
  for rel_filename in find_all_py_source_files():
    print("travis_fold:start:pylint.%s" % rel_filename)
    extra_args = []
    if "/" in rel_filename and not rel_filename.startswith("returnn/"):  # demos or tools
      extra_args += [r"--module-rgx=([a-z_][a-z0-9_\-]*)$"]
      extra_args += ["--disable=wrong-import-position"]  # for now
    proc = subprocess.Popen(["pylint", rel_filename] + extra_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, _ = proc.communicate()
    stdout = stdout.decode("utf8")
    file_has_errors = proc.returncode != 0
    print(color.color(
      "File: %s" % rel_filename,
      color="black" if (rel_filename in ignore_count_for_files or not file_has_errors) else "red"))
    if "EXCEPTION" in stdout and "RecursionError" in stdout:
      # https://github.com/PyCQA/pylint/issues/1452
      # https://github.com/PyCQA/astroid/issues/437
      # Don't print full stdout. It will spam too much.
      print("PyLint issue #1452 triggered. https://github.com/PyCQA/pylint/issues/1452")
    elif rel_filename in ignore_count_for_files:
      print(stdout[:1000])
      print("... (ignored further output; file is not relevant)")
    else:
      print(stdout)
    print("Return code:", proc.returncode)
    if file_has_errors:
      if rel_filename in ignore_count_for_files:
        print("The inspection reports for this file are currently ignored.")
      else:
        print(color.color("The inspection reports for this file are fatal!", color="red"))
        num_relevant_files_with_errors += 1
    print("travis_fold:end:pylint.%s" % rel_filename)
  if num_relevant_files_with_errors:
    print("Num relevant files with errors:", num_relevant_files_with_errors)
    sys.exit(1)


if __name__ == "__main__":
  better_exchook.install()
  main()




import sys
import os

import _setup_test_env  # noqa
import returnn.theano.native_op as theano_native_op
import returnn.native_op as native_op
import numpy
from numpy.testing.utils import assert_almost_equal
import theano.tensor as T
import returnn.theano.util as theano_util
import sys
f32 = "float32"


theano_util.monkey_patches()


chunk = theano_native_op.chunk
unchunk = theano_native_op.unchunk
naive_chunk_start_frames = native_op.Chunking.naive_chunk_start_frames


def get_num_chunks(n_time, chunk_size, chunk_step):
  return len(naive_chunk_start_frames(n_time, chunk_size, chunk_step))


def naive_chunk(x, chunk_size, chunk_step):
  if x.ndim == 3:  # (time,batch,dim)
    if x.shape[1] == 1:
      return naive_chunk(x[:, 0], chunk_size, chunk_step)
    raise NotImplemented
  assert x.ndim == 2  # (time,dim)
  chunk_start_frames = naive_chunk_start_frames(x.shape[0], chunk_size, chunk_step)
  n_chunks = len(chunk_start_frames)
  out = numpy.zeros((chunk_size, n_chunks, x.shape[1]), dtype=x.dtype)
  oindex = numpy.zeros((chunk_size, n_chunks), dtype="int8")
  for b, t in enumerate(chunk_start_frames):
    t2 = min(t + chunk_size, x.shape[0])
    out[:t2 - t, b] = x[t:t2]
    oindex[:t2 - t, b] = 1
  return out, oindex


def test_chunk():
  n_time = 101
  n_batch = 1
  n_dim = 5
  chunk_size = 11
  chunk_step = 7
  numpy.random.seed(1234)
  _x = numpy.random.randn(n_time, n_batch, n_dim).astype(f32)
  _out, _oindex = naive_chunk(_x, chunk_size, chunk_step)
  _index = numpy.ones((n_time, n_batch), dtype="int8")
  x = T.as_tensor(_x)
  index = T.as_tensor(_index)
  out, oindex = chunk(x, index=index, chunk_size=chunk_size, chunk_step=chunk_step)
  _out2 = out.eval()
  _oindex2 = oindex.eval()
  assert _out.shape == _out2.shape
  assert _oindex.shape == _oindex2.shape
  assert_almost_equal(_oindex, _oindex2)
  assert_almost_equal(_out, _out2)


def test_chunk_unchunk():
  n_time = 101
  n_batch = 3
  n_dim = 5
  _x = numpy.random.randn(n_time, n_batch, n_dim).astype(f32)
  _index = numpy.ones((n_time, n_batch), dtype="int8")
  x = T.as_tensor(_x)
  index = T.as_tensor(_index)
  chunk_size = 11
  chunk_step = 7
  n_chunks = get_num_chunks(n_time, chunk_size, chunk_step)
  #print >>sys.stderr, "Should have %i n_chunks." % n_chunks
  out, oindex = chunk(x, index=index, chunk_size=chunk_size, chunk_step=chunk_step)
  x2, index2, factors = unchunk(out, index=oindex, chunk_size=chunk_size, chunk_step=chunk_step, n_time=x.shape[0], n_batch=x.shape[1])
  _x2 = x2.eval()
  _index2 = index2.eval()
  assert_almost_equal(_x, _x2)
  assert numpy.all(_index2)


def test_chunk_unchunk_grad():
  n_time = 101
  n_batch = 3
  n_dim = 5
  numpy.random.seed(1234)
  _x = numpy.random.randn(n_time, n_batch, n_dim).astype(f32)
  _index = numpy.ones((n_time, n_batch), dtype="int8")
  x = T.as_tensor(_x)
  index = T.as_tensor(_index)
  chunk_size = 11
  chunk_step = 7
  out, oindex = chunk(x, index=index, chunk_size=chunk_size, chunk_step=chunk_step)
  x2, index2, factors = unchunk(out, index=oindex, chunk_size=chunk_size, chunk_step=chunk_step, n_time=x.shape[0], n_batch=x.shape[1])
  grad = T.grad(T.sum((x2 - x) ** 2), wrt=x)
  _grad = grad.eval()
  assert_almost_equal(_grad, 0)


def test_chunk_unchunk_grad2():
  n_time = 101
  n_batch = 3
  n_dim = 5
  numpy.random.seed(1234)
  _x = numpy.random.randn(n_time, n_batch, n_dim).astype(f32)
  _Dx2 = numpy.random.randn(n_time, n_batch, n_dim).astype(f32)
  _index = numpy.ones((n_time, n_batch), dtype="int8")
  x = T.as_tensor(_x)
  Dx2 = T.as_tensor(_Dx2)
  index = T.as_tensor(_index)
  chunk_size = 11
  chunk_step = 7

  out, oindex = chunk(x, index=index, chunk_size=chunk_size, chunk_step=chunk_step)
  chunk_op = native_op.Chunking().make_theano_op()
  assert type(out.owner.op) is type(chunk_op)

  x2, index2, factors = unchunk(out, index=oindex, chunk_size=chunk_size, chunk_step=chunk_step, n_time=x.shape[0], n_batch=x.shape[1])
  unchunk_op = native_op.UnChunking().make_theano_op()
  assert type(x2.owner.op) is type(unchunk_op)

  Dout, _, _, _, _, _ = unchunk_op.grad(x2.owner.inputs, (Dx2, None, None))
  Dx, _, _, _, _ = chunk_op.grad(out.owner.inputs, (Dout, None))
  _Dx = Dx.eval()
  assert_almost_equal(_Dx, _Dx2)



#!/usr/bin/env python3

"""
This demonstrates how to use RETURNN as a framework.
"""

from __future__ import print_function

import tensorflow as tf

import returnn  # pip install returnn
from returnn.tf.engine import Engine
from returnn.datasets import init_dataset
from returnn.config import get_global_config
from returnn.util.basic import get_login_username
from returnn.util import better_exchook


print("TF version:", tf.__version__)
print("RETURNN imported from:", returnn.__file__)
print("RETURNN version:", returnn.__version__)
print("RETURNN long version:", returnn.__long_version__)

better_exchook.install()


config = get_global_config(auto_create=True)
config.update(dict(
  batching="random",
  batch_size=5000,
  max_seqs=10,
  chunking="0",

  network={
    "fw0": {"class": "rec", "unit": "NativeLstm2", "dropout": 0.1, "n_out": 10, "from": "data:data"},
    "output": {"class": "softmax", "loss": "ce", "from": "fw0"}
  },

  # training
  optimizer={'class': 'adam'},
  learning_rate=0.01,
  num_epochs=5,
  debug_add_check_numerics_ops=True,

  model="/tmp/%s/returnn-demo-as-framework/model" % get_login_username(),
  cleanup_old_models=True,

  learning_rate_control="newbob_multi_epoch",
  learning_rate_control_relative_error_relative_lr=True,
  newbob_multi_num_epochs=3, newbob_multi_update_interval=1, newbob_learning_rate_decay=0.9,
  learning_rate_file="/tmp/%s/returnn-demo-as-framework/newbob.data" % get_login_username(),

  # log
  log_verbosity=3
))

engine = Engine(config)

train_data = init_dataset({"class": "Task12AXDataset", "num_seqs": 1000, "name": "train"})
dev_data = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "dev", "fixed_random_seed": 1})

engine.init_train_from_config(train_data=train_data, dev_data=dev_data)
engine.train()




"""
This defines the base dataset class :class:`Dataset`.
"""

from __future__ import print_function

__author__ = "Patrick Doetsch"
__copyright__ = "Copyright 2015"
__credits__ = ["Patrick Doetsch", "Paul Voigtlaender"]
__license__ = "RWTHASR"
__version__ = "0.9"
__maintainer__ = "Patrick Doetsch"
__email__ = "doetsch@i6.informatik.rwth-aachen.de"

from threading import RLock
from random import Random, random

import sys
import os
import numpy
import functools
import typing

from returnn.log import log
from returnn.engine.batch import Batch, BatchSetGenerator
from returnn.datasets.util.vocabulary import Vocabulary
from returnn.util.basic import try_run, NumbersDict, unicode, OptionalNotImplementedError


class Dataset(object):
  """
  Base class for any dataset. This defines the dataset API.
  """

  @staticmethod
  def kwargs_update_from_config(config, kwargs):
    """
    :type config: Config.Config
    :type kwargs: dict[str]
    """
    def set_or_remove(key, value):
      """
      :param str key:
      :param value:
      """
      if key in kwargs and kwargs[key] is None:
        del kwargs[key]
      if value is not None and key not in kwargs:
        kwargs[key] = value

    set_or_remove("window", config.int('window', 0) or None)
    set_or_remove("context_window", config.typed_value("context_window"))
    set_or_remove("chunking", config.opt_typed_value("chunking", None))
    set_or_remove("seq_ordering", config.value("batching", None))
    set_or_remove("shuffle_frames_of_nseqs", config.int('shuffle_frames_of_nseqs', 0) or None)
    set_or_remove("min_chunk_size", config.int('min_chunk_size', 0) or None)
    set_or_remove("chunking_variance", config.float("chunking_variance", 0))

  @staticmethod
  def get_default_kwargs_eval(config):
    """
    :param Config.Config config:
    :rtype: dict[str]
    """
    # For dev/eval, by default, we should not do chunking (i.e. chunking = "0").
    chunking = "0"
    if config.value("on_size_limit", "ignore") == "chunk":
      chunking = config.value("batch_size", "0")
    elif config.value('chunking', "0") == "1":  # MLP mode
      chunking = "1"
    elif config.bool('chunk_eval', False):
      chunking = config.value('chunking', "0")
    return dict(chunking=chunking, seq_ordering="sorted", shuffle_frames_of_nseqs=0)

  @classmethod
  def from_config(cls, config, **kwargs):
    """
    :type config: Config.Config
    :param dict[str] kwargs: passed on to __init__
    :rtype: Dataset
    """
    cls.kwargs_update_from_config(config, kwargs)
    return cls(**kwargs)

  def __init__(self, name=None,
               window=1, context_window=None, chunking=None,
               seq_ordering='default', random_seed_offset=None,
               partition_epoch=None, repeat_epoch=None,
               seq_list_filter_file=None, unique_seq_tags=False,
               seq_order_seq_lens_file=None,
               shuffle_frames_of_nseqs=0, min_chunk_size=0, chunking_variance=0,
               estimated_num_seqs=None):
    """
    :param str name: e.g. "train" or "eval"
    :param int window: features will be of dimension window * feature_dim, as we add a context-window around.
      not all datasets support this option.
    :param None|int|dict|NumbersDict|(dict,dict) context_window: will add this context for each chunk
    :param None|str|int|(int,int)|dict|(dict,dict)|function chunking: "chunk_size:chunk_step"
    :param str seq_ordering: "batching"-option in config. e.g. "default", "sorted" or "random".
      See self.get_seq_order_for_epoch() for more details.
    :param int|None random_seed_offset:
    :param int|None partition_epoch:
    :param int|None repeat_epoch: Repeat the sequences in an epoch this many times. Useful to scale the dataset
      relative to other datasets, e.g. when used in CombinedDataset. Not allowed to be used in combination with
      partition_epoch.
    :param str|None seq_list_filter_file: defines a subset of sequences (by tag) to use
    :param bool unique_seq_tags: uniquify seqs with same seq tags in seq order
    :param str|None seq_order_seq_lens_file: for seq order, use the seq length given by this file
    :param int shuffle_frames_of_nseqs: shuffles the frames. not always supported
    :param None|int estimated_num_seqs: for progress reporting in case the real num_seqs is unknown
    """
    self.name = name or ("dataset_id%s" % id(self))
    self.lock = RLock()  # Used when manipulating our data potentially from multiple threads.
    self.rnd_seq_drop = None  # type: typing.Optional[Random]
    self.num_inputs = 0  # usually not used, but num_outputs instead, which is more generic
    self.num_outputs = None  # type: typing.Optional[typing.Dict[str,typing.Tuple[int,int]]]  # tuple is num-classes, len(shape).  # nopep8
    self.window = window
    self.seq_ordering = seq_ordering  # "default", "sorted" or "random". See self.get_seq_order_for_epoch().
    if random_seed_offset is None:
      random_seed_offset = self._get_default_random_seed_offset()
    self.random_seed_offset = random_seed_offset
    self.partition_epoch = partition_epoch or 1
    self.repeat_epoch = repeat_epoch or 1
    self.seq_tags_filter = set(self._load_seq_list_file(seq_list_filter_file)) if seq_list_filter_file else None
    self.unique_seq_tags = unique_seq_tags
    self._seq_order_seq_lens_file = seq_order_seq_lens_file
    self._seq_order_seq_lens_by_idx = None
    # There is probably no use case for combining the two, so avoid potential misconfiguration.
    assert self.partition_epoch == 1 or self.repeat_epoch == 1, (
      "Combining partition_epoch and repeat_epoch is prohibited.")
    self.labels = {}  # type: typing.Dict[str,typing.List[str]]
    self.weights = {}
    self._num_timesteps = 0
    self._num_seqs = 0
    self._estimated_num_seqs = estimated_num_seqs
    self.min_chunk_size = min_chunk_size
    self.chunking_variance = chunking_variance
    self.chunk_size, self.chunk_step, self.custom_chunking_func = self._parse_chunking(chunking)
    if isinstance(context_window, (tuple, list)):
      assert len(context_window) == 2
      for elem in context_window:
        assert isinstance(elem, dict)
      # assume that first element of tuple|list is for left context and second element for right context
      self.ctx_left = NumbersDict(numbers_dict=context_window[0])
      self.ctx_right = NumbersDict(numbers_dict=context_window[1])
    else:
      if context_window is None:
        context_window = NumbersDict()
      elif isinstance(context_window, int):
        context_window = NumbersDict(numbers_dict={"data": context_window})
      elif isinstance(context_window, dict):
        context_window = NumbersDict(numbers_dict=context_window)
      assert isinstance(context_window, NumbersDict)
      # ctx_total is how much frames we add additionally.
      # One less because the original frame also counts, and context_window=1 means that we just have that single frame.
      ctx_total = NumbersDict.max([context_window, 1]) - 1
      # In case ctx_total is odd / context_window is even, we have to decide where to put one more frame.
      # To keep it consistent with e.g. 1D convolution with a kernel of even size, we add one more to the right.
      # See test_tfconv1d_evensize().
      self.ctx_left = ctx_total // 2
      self.ctx_right = ctx_total - self.ctx_left
    assert isinstance(self.ctx_left, NumbersDict)
    assert isinstance(self.ctx_right, NumbersDict)
    self.shuffle_frames_of_nseqs = shuffle_frames_of_nseqs
    self.epoch = None

  def __repr__(self):
    return "<%s %r epoch=%s>" % (
      self.__class__.__name__,
      getattr(self, "name", "<unknown>"),
      getattr(self, "epoch", "<unknown>"))

  @staticmethod
  def _get_default_random_seed_offset():
    """
    :return: 0 usually
    :rtype: int
    """
    from returnn.config import get_global_config
    config = get_global_config(raise_exception=False)
    if not config:
      return 0
    if config.is_true("use_horovod"):
      import returnn.tf.horovod
      if returnn.tf.horovod.get_ctx().is_dataset_distribution_random_seed_offset():
        return returnn.tf.horovod.get_ctx().rank() * 16127
    return 0

  @staticmethod
  def _parse_chunking(chunking):
    """
    :param None|str|int|(int,int)|dict|(dict,dict)|(NumbersDict,NumbersDict)|function chunking:
      as it comes from the config / from the user
    :return: chunk_size, chunk_step, custom_chunking_func
    :rtype: (NumbersDict|None,NumbersDict|None,function|None)
    """
    if callable(chunking):
      return None, None, chunking
    if isinstance(chunking, str):
      if ":" in chunking:
        chunking = tuple(map(int, chunking.split(":")))
      else:
        chunking = int(chunking)
    if not isinstance(chunking, (tuple, list)):
      chunking = (chunking, None)
    chunk_size, chunk_step = chunking
    if chunk_size is None:
      chunk_size = 0
    assert isinstance(chunk_size, (int, dict, NumbersDict))
    chunk_size = NumbersDict(chunk_size)
    assert chunk_size == 0 or chunk_size.min_value() > 0, "chunk size must not be negative"
    if chunk_step in (None, 0):
      chunk_step = chunk_size
    assert isinstance(chunk_step, (int, dict, NumbersDict))
    chunk_step = NumbersDict(chunk_step)
    if chunk_size != 0:
      assert sorted(chunk_step.keys()) == sorted(chunk_size.keys())
      assert chunk_step.max_value() > 0, "chunking step must be positive (for some key)"
    return chunk_size, chunk_step, None

  @staticmethod
  def _load_seq_list_file(filename, use_cache_manager=False, expect_list=True):
    """
    :param str filename:
    :param bool use_cache_manager:
    :param bool expect_list:
    :rtype: list[str]|dict[str,list[str]]
    """
    if use_cache_manager:
      import returnn.util.basic
      filename = returnn.util.basic.cf(filename)
    if filename.endswith(".pkl"):
      import pickle
      seq_list = pickle.load(open(filename, 'rb'))
      if expect_list:
        assert isinstance(seq_list, list)
    elif filename.endswith(".gz"):
      import gzip
      seq_list = gzip.open(filename, "rt").read().splitlines()
    else:
      seq_list = open(filename).read().splitlines()
    return seq_list

  def _sliding_window(self, xr):
    """
    :type xr: numpy.ndarray
    :rtype: numpy.ndarray
    """
    # noinspection PyProtectedMember
    from numpy.lib.stride_tricks import as_strided
    x = numpy.concatenate([self.zpad, xr, self.zpad])
    return as_strided(
      x,
      shape=(x.shape[0] - self.window + 1, 1, self.window, self.num_inputs),
      strides=(x.strides[0], x.strides[1] * self.num_inputs) + x.strides
      ).reshape((xr.shape[0], self.num_inputs * self.window))

  def is_cached(self, start, end):
    """
    :param int start: like in load_seqs(), sorted seq idx
    :param int end: like in load_seqs(), sorted seq idx
    :rtype: bool
    :returns whether we have the full range (start,end) of sorted seq idx.
    """
    if start == end:
      return True  # Empty.
    assert start < end
    return False

  def get_seq_length(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: NumbersDict
    :returns the len of the input features and the len of the target sequence.
    """
    raise NotImplementedError

  def get_estimated_seq_length(self, seq_idx):
    """
    In contrast to self.get_seq_length(), this method is designed to work for sequences that have not been loaded yet
    via self.load_seqs().
    Used by meta-datasets for sequence ordering. Currently we only provide one number, i.e. do not give different
    estimates for the different data keys (as in get_seq_length()). It is up to the dataset what this number represents
    and how it is computed.

    :param int seq_idx: for current epoch, not the corpus seq idx
    :rtype: int
    :returns sequence length estimate (for sorting)
    """
    raise OptionalNotImplementedError

  def get_num_timesteps(self):
    """
    :rtype: int
    """
    assert self._num_timesteps > 0
    return self._num_timesteps

  def load_seqs(self, start, end):
    """
    Load data sequences, such that self.get_data() & friends can return the data.

    :param int start: start sorted seq idx, inclusive
    :param int end: end sorted seq idx, exclusive
    """
    assert start >= 0
    assert start <= end
    if self.is_cached(start, end):
      return

    if self.shuffle_frames_of_nseqs > 0:
      # We always load N seqs at once and shuffle all their frames.
      start, end = self._get_load_seqs_superset(start, end)
      self._load_seqs(start, end)
      while start < end:
        self._shuffle_frames_in_seqs(start, start + self.shuffle_frames_of_nseqs)
        start += self.shuffle_frames_of_nseqs
    else:
      self._load_seqs(start, end)

  def _get_load_seqs_superset(self, start, end):
    """
    :type start: int
    :type end: int
    :returns the superset (start,end) of seqs to be loaded.
    For shuffle_frames_of_nseqs > 0, we always load N seqs at once
    and shuffle all their frames,
    thus start/end will be aligned to self.shuffle_frames_of_nseqs.
    """
    assert start <= end
    assert start < self.num_seqs
    if self.shuffle_frames_of_nseqs > 0:
      m = self.shuffle_frames_of_nseqs
      start -= start % m
      end += (m - (end % m)) % m
    return start, end

  def _shuffle_frames_in_seqs(self, start, end):
    raise OptionalNotImplementedError

  def _load_seqs(self, start, end):
    """
    Load data sequences.
    :param int start: inclusive seq idx start
    :param int end: exclusive seq idx end. can be more than num_seqs
    If end > num_seqs, will not load them.
    """
    raise NotImplementedError

  def _get_seq_order_seq_lens_by_idx(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: int
    """
    if not self._seq_order_seq_lens_by_idx:
      assert self._seq_order_seq_lens_file
      if self._seq_order_seq_lens_file.endswith(".gz"):
        import gzip
        raw = gzip.GzipFile(self._seq_order_seq_lens_file, "rb").read()
      else:
        raw = open(self._seq_order_seq_lens_file, "rb").read()
      seq_lens = eval(raw)
      assert isinstance(seq_lens, dict)
      all_tags = self.get_all_tags()
      self._seq_order_seq_lens_by_idx = [seq_lens[tag] for tag in all_tags]
    return self._seq_order_seq_lens_by_idx[seq_idx]

  def get_seq_order_for_epoch(self, epoch, num_seqs, get_seq_len=None):
    """
    Returns the order of the given epoch.
    This is mostly a static method, except that is depends on the configured type of ordering,
    such as 'default' (= as-is), 'sorted' or 'random'. 'sorted' also uses the sequence length.

    :param int epoch: for 'random', this determines the random seed
    :param int num_seqs:
    :param ((int) -> int)|None get_seq_len: function (originalSeqIdx: int) -> int
    :return: the order for the given epoch. such that seq_idx -> underlying idx
    :rtype: typing.Sequence[int]
    """
    partition_epoch = self.partition_epoch or 1
    repeat_epoch = self.repeat_epoch or 1
    assert num_seqs > 0
    if self._seq_order_seq_lens_file:
      get_seq_len = self._get_seq_order_seq_lens_by_idx

    if self.seq_ordering == 'default':
      seq_index = range(num_seqs)
    elif self.seq_ordering.startswith("default_every_n:"):
      # This order is useful if you have "initial_state": "keep_over_epoch",
      # where num == max_seqs, batch_size = inf, max_seq_len = inf, chunking = None.
      _, num = self.seq_ordering.split(":")
      num = int(num)
      seq_index = numpy.arange(num_seqs // num, dtype="int64").repeat(num)
      for i in range(1, num):
        seq_index[i::num] += i * (num_seqs // num)
    elif self.seq_ordering == 'reverse':
      seq_index = range(num_seqs - 1, -1, -1)
    elif self.seq_ordering in ['sorted', 'sorted_reverse']:
      assert get_seq_len
      reverse = -1 if self.seq_ordering == 'sorted_reverse' else 1
      seq_lens = [reverse * get_seq_len(i) for i in range(num_seqs)]
      seq_index = numpy.argsort(seq_lens, kind="stable")
    elif self.seq_ordering.startswith('random'):
      tmp = self.seq_ordering.split(':')
      nth = int(tmp[1]) if len(tmp) > 1 else 1
      # Keep this deterministic! Use fixed seed.
      rnd_seed = self._get_random_seed_for_epoch(epoch=epoch, num_epochs_fixed=nth)
      random_generator = numpy.random.RandomState(rnd_seed)
      seq_index = random_generator.permutation(num_seqs)
    elif self.seq_ordering.startswith('sort_bin_shuffle'):
      # Shuffle seqs, sort by length, and shuffle bins (then shuffle seqs within each bin if sort_bin_shuffle_x2).
      assert get_seq_len
      tmp = self.seq_ordering.split(':')[1:]
      # Keep this deterministic! Use fixed seed.
      if len(tmp) <= 1:
        nth = 1
      else:
        nth = int(tmp[1])
      rnd_seed = self._get_random_seed_for_epoch(epoch=epoch, num_epochs_fixed=nth)
      random_generator = numpy.random.RandomState(rnd_seed)
      seq_index = random_generator.permutation(num_seqs).tolist()  # type: typing.List[int]
      seq_index.sort(key=get_seq_len)  # Sort by length, starting with shortest.
      if len(tmp) == 0:
        bins = 2
      else:
        if tmp[0].startswith("."):  # starting with "." -> approx chunk size (num of seqs in one bin)
          bins = max(num_seqs // int(tmp[0][1:]), 2)
        else:  # the number of bins
          bins = int(tmp[0])
      bin_ids = random_generator.permutation(bins)  # Shuffle bins.
      out_index = []
      for i in bin_ids:
        if i == bins - 1:
          part = seq_index[i * len(seq_index) // bins:][:]
        else:
          part = seq_index[i * len(seq_index) // bins:(i + 1) * len(seq_index) // bins][:]
        if self.seq_ordering.startswith('sort_bin_shuffle_x2'):
          random_generator.shuffle(part)  # Shuffle within the bin.
        out_index.append(part)
      seq_index = numpy.concatenate(out_index)
    elif self.seq_ordering.startswith('laplace'):
      assert get_seq_len
      tmp = self.seq_ordering.split(':')[1:]
      if len(tmp) == 0:
        bins = 2
      else:
        if tmp[0].startswith("."):  # starting with "." -> approx chunk size (num of seqs in one bin)
          bins = max(num_seqs // int(tmp[0][1:]), 2)
        else:  # the number of bins
          bins = int(tmp[0])
      if len(tmp) <= 1:
        nth = 1
      else:
        nth = int(tmp[1])
      rnd_seed = self._get_random_seed_for_epoch(epoch=epoch, num_epochs_fixed=nth)
      random_generator = numpy.random.RandomState(rnd_seed)
      seq_index = random_generator.permutation(num_seqs)  # type: numpy.ndarray
      out_index = []
      for i in range(bins):
        if i == bins - 1:
          part = seq_index[i * len(seq_index) // bins:].tolist()
        else:
          part = seq_index[i * len(seq_index) // bins:(i + 1) * len(seq_index) // bins].tolist()
        part.sort(key=get_seq_len, reverse=(i % 2 == 1))
        out_index += part
      seq_index = out_index
    else:
      assert False, "invalid batching specified: " + self.seq_ordering

    if self.unique_seq_tags:
      # Note: This is as generic as possible, but requires that get_all_tags is implemented.
      all_seq_tags = self.get_all_tags()
      used_seq_tags = set()
      seq_index = [
        i for i in seq_index
        if (all_seq_tags[i] not in used_seq_tags, used_seq_tags.add(all_seq_tags[i]))[0]]
    if partition_epoch > 1:
      seq_index = self._apply_partition_epoch(seq_index, partition_epoch, epoch)
    if repeat_epoch > 1:
      seq_index = list(seq_index) * repeat_epoch
    if self.seq_tags_filter is not None:
      # Note: This is as generic as possible, but requires that get_all_tags is implemented.
      assert seq_index
      all_seq_tags = self.get_all_tags()
      assert len(all_seq_tags) == num_seqs == self.get_total_num_seqs(), "%r vs %r vs %r" % (
        len(all_seq_tags), num_seqs, self.get_total_num_seqs())
      old_seq_index = seq_index
      seq_index = [i for i in seq_index if all_seq_tags[i] in self.seq_tags_filter]
      assert seq_index, "%s: empty after applying seq_list_filter_file. Example filter tags: %r, used tags: %r" % (
        self, sorted(self.seq_tags_filter)[:3], [all_seq_tags[i] for i in old_seq_index[:3]])
    return seq_index

  @classmethod
  def _apply_partition_epoch(cls, seq_index, partition_epoch, epoch):
    """
    :param typing.Sequence[int] seq_index: full list of ordered sequence indices
    :param int partition_epoch: number of partitions seq_index should be split into
    :param int|None epoch: current epoch
    :return: partition of seq_index for current epoch
    :rtype: typing.Sequence[int]
    """
    num_seqs = len(seq_index)
    current_partition = ((epoch or 1) - 1) % partition_epoch
    seqs_per_epoch = num_seqs // partition_epoch
    partition_sizes = ([seqs_per_epoch + 1] * (num_seqs % partition_epoch) +
                       [seqs_per_epoch] * (partition_epoch - num_seqs % partition_epoch))
    assert sum(partition_sizes) == num_seqs and len(partition_sizes) == partition_epoch
    partitions = functools.reduce(lambda a, x: a + [a[-1] + x], partition_sizes, [0])  # cumulative sum
    assert len(partitions) == partition_epoch + 1
    seq_index = seq_index[partitions[current_partition]:partitions[current_partition + 1]]
    assert len(seq_index) == partition_sizes[current_partition]

    return seq_index

  def _get_random_seed_for_epoch(self, epoch, num_epochs_fixed=1):
    """
    :param int|None epoch:
    :param int num_epochs_fixed: keep random seed fixed for n subsequent epochs
    :rtype: int
    """
    partition_epoch = self.partition_epoch or 1
    seed = epoch or 1
    if partition_epoch > 1:
      seed = (seed - 1) // partition_epoch + 1  # taking partitions requires constant seed during full epoch
    if num_epochs_fixed > 1:
      seed = (seed - 1) // num_epochs_fixed + 1
    return seed + self.random_seed_offset

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :type epoch: int|None
    :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
    :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order. Only possible
      if the dataset has such indices (see self.have_corpus_seq_idx()).
    :rtype: bool
    :returns whether the order changed (True is always safe to return)

    This is called when we start a new epoch, or at initialization.
    Call this when you reset the seq list.
    """
    self.epoch = epoch
    self.rnd_seq_drop = Random(self._get_random_seed_for_epoch(epoch=epoch))
    return False

  def finish_epoch(self):
    """
    This would get called at the end of the epoch (currently optional only).
    After this, further calls to :func:`get_data` or :func:`load_seqs` are invalid,
    until a new call to :func:`init_seq_order` follows.
    """
    self.epoch = None

  def get_current_seq_order(self):
    """
    :return: many datasets use self.get_seq_order_for_epoch. this function would return the current seq order
      for the current epoch, after self.init_seq_order was called.
      Not all datasets implement this.
    :rtype: typing.Sequence[int]
    """
    raise OptionalNotImplementedError

  def _base_init(self):
    self.zpad = None
    # We expect that the following attributes are already set elsewhere, by a derived class.
    assert self.num_outputs
    if not self.num_inputs:
      assert not self.window or self.window in (0, 1) or "data" in self.num_outputs
      return
    assert self.num_inputs > 0
    assert self.window > 0

    if int(self.window) % 2 == 0:
      self.window += 1

    if self.window > 1:
      self.zpad = numpy.zeros((int(self.window) // 2, self.num_inputs), dtype=numpy.float32)

  def initialize(self):
    """
    Does the main initialization before it can be used.
    This needs to be called before self.load_seqs() can be used.
    """
    self._base_init()
    self.init_seq_order()

  def get_times(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    """
    raise OptionalNotImplementedError

  def get_data(self, seq_idx, key):
    """
    :param int seq_idx: sorted seq idx
    :param str key: data-key, e.g. "data" or "classes"
    :rtype: numpy.ndarray
    :returns features or targets: format 2d (time,feature) (float)
    """
    # Fallback implementation for old-style subclasses.
    if key == "data":
      return self.get_input_data(seq_idx)
    else:
      return self.get_targets(key, seq_idx)

  def get_input_data(self, sorted_seq_idx):
    """
    :type sorted_seq_idx: int
    :rtype: numpy.ndarray
    :returns features: format 2d (time,feature) (float)
    """
    raise NotImplementedError

  def get_targets(self, target, sorted_seq_idx):
    """
    :param str target: data key
    :type sorted_seq_idx: int
    :rtype: numpy.ndarray
    :returns targets: format 1d (time) (int: idx of output-feature)
    """
    # For new-style subclasses, which just provide get_data.
    return self.get_data(sorted_seq_idx, target)

  def get_ctc_targets(self, sorted_seq_idx):
    """
    Warning: This is deprecated/obsolete.

    :param int sorted_seq_idx:
    :rtype: numpy.ndarray|None
    """
    return None

  def get_data_slice(self, seq_idx, key, start_frame, end_frame):
    """
    :param int seq_idx:
    :param str key:
    :param int start_frame:
    :param int end_frame:
    :return: x[start_frame:end_frame], with x = get_data(seq_idx, key)
    :rtype: numpy.ndarray
    """
    if "[sparse:" in key and (start_frame > 0 or end_frame < self.get_seq_length(seq_idx)[key]):
      return self._get_data_slice_sparse(seq_idx, key, start_frame, end_frame)
    data = self.get_data(seq_idx, key)
    return data[start_frame:end_frame]

  def _get_data_slice_sparse(self, seq_idx, key, start_frame, end_frame):
    key_prefix = key[:key.index("[")]
    sparse_info = key[key.index("[") + 1:key.index("]")].split(":")
    assert len(sparse_info) == 4
    assert tuple(sparse_info[0:3]) == ("sparse", "coo", "2")
    s0 = self.get_data(seq_idx, "%s[sparse:coo:2:0]" % key_prefix)
    assert s0 is not None
    from returnn.native_op import sparse_splice_offset_numpy
    s0_start = sparse_splice_offset_numpy(s0, start_frame)
    s0_end = sparse_splice_offset_numpy(s0, end_frame)
    if sparse_info[-1] == "0":
      return s0[s0_start:s0_end] - start_frame
    else:
      data = self.get_data(seq_idx, key)
      return data[s0_start:s0_end]

  def get_tag(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    :rtype: str
    """
    return "seq-%i" % sorted_seq_idx

  def get_all_tags(self):
    """
    :return: list of all seq tags, of the whole dataset, without partition epoch.
      Note that this is not possible with all datasets.
    :rtype: list[str]
    """
    old_partition_epoch = self.partition_epoch
    try:
      all_tags = [None] * self.num_seqs  # type: typing.List[typing.Union[None,str]]
      for seq_idx in range(self.num_seqs):
        all_tags[seq_idx] = self.get_tag(seq_idx)
      return all_tags
    finally:
      self.partition_epoch = old_partition_epoch

  def get_total_num_seqs(self):
    """
    :return: total number of seqs, without partition epoch.
      Should be the same as len(self.get_all_tags()).
      Note that this is not possible with all datasets.
    :rtype: int
    """
    if self.partition_epoch == 1:
      # Note: self.num_seqs might not always be set, or even be correct...
      return self.num_seqs
    raise NotImplementedError("%s: get_total_num_seqs with partition epoch %i" % (self, self.partition_epoch))

  def have_corpus_seq_idx(self):
    """
    :rtype: bool
    :return: whether you can call self.get_corpus_seq_idx()
    """
    return False

  def get_corpus_seq_idx(self, seq_idx):
    """
    :param int seq_idx: sorted sequence index from the current epoch, depending on seq_ordering
    :return: the sequence index as-is in the original corpus (as if you would have sorting="default").
      only defined if self.have_corpus_seq_idx()
    :rtype: int
    """
    if self.seq_ordering == "default":
      return seq_idx
    assert self.have_corpus_seq_idx()
    raise NotImplemented

  def has_ctc_targets(self):
    """
    :return: whether we have get_ctc_targets implemented
    :rtype: bool
    """
    return False

  @classmethod
  def generic_complete_frac(cls, seq_idx, num_seqs):
    """
    :param int seq_idx: idx
    :param int|None num_seqs: None if not available
    :return: Returns a fraction (float in [0,1], always > 0) of how far we have advanced
      for this seq in the dataset.
      This does not have to be exact. This is only for the user.
    """
    if num_seqs:
      return min(float(seq_idx + 1) / num_seqs, 1.0)
    else:
      # We don't know. So:
      # Some monotonic increasing function in [0,1] which never reaches 1.
      import math
      return max(1.e-10, 1.0 - math.exp(-seq_idx * 1000))

  def get_complete_frac(self, seq_idx):
    """
    :param int seq_idx:
    :return: Returns a fraction (float in [0,1], always > 0) of how far we have advanced
      for this seq in the dataset.
      This does not have to be exact. This is only for the user.
    :rtype: float
    """
    # noinspection PyBroadException
    try:
      num_seqs = self.num_seqs
    except Exception:  # num_seqs not always available
      # noinspection PyBroadException
      try:
        num_seqs = self.estimated_num_seqs
      except Exception:  # also not always available
        num_seqs = None  # ignore
    return self.generic_complete_frac(seq_idx, num_seqs)

  @property
  def num_seqs(self):
    """
    :rtype: int
    """
    raise NotImplementedError

  @property
  def estimated_num_seqs(self):
    """
    :return: estimated num seqs. does not have to be exact
    :rtype: int|None
    """
    # noinspection PyBroadException
    try:
      return self.num_seqs
    except Exception:  # might not be available
      pass
    if self._estimated_num_seqs is not None:
      return self._estimated_num_seqs
    return None

  def get_data_keys(self):
    """
    :return: all available data keys (for get_data and all other functions)
    :rtype: list[str]
    """
    return ["data"] + self.get_target_list()

  def get_target_list(self):
    """
    :return: subset of :func:`get_data_keys`. target keys are usually not available during inference
    :rtype: list[str]
    """
    return ["classes"]

  def get_data_dim(self, key):
    """
    :param str key: e.g. "data" or "classes"
    :return: number of classes, no matter if sparse or not
    :rtype: int
    """
    if key in self.num_outputs:
      # num_outputs should have the correct dimension, even for key "data" with self.window > 1.
      return self.num_outputs[key][0]
    if self.window > 1 and key == "data":
      assert self.num_inputs
      return self.num_inputs * self.window
    return 1  # unknown

  def get_data_dtype(self, key):
    """
    :param str key: e.g. "data" or "classes"
    :return: dtype as str, e.g. "int32" or "float32"
    :rtype: str
    """
    if self.is_data_sparse(key):
      return "int32"
    return "float32"

  def is_data_sparse(self, key):
    """
    :param str key: e.g. "data" or "classes"
    :return: whether the data is sparse
    :rtype: bool
    """
    # Note: We cannot call get_data_dtype, as we would maybe result in infinite recursion.
    if key in self.num_outputs:
      return self.num_outputs[key][1] <= 1
    if key == "data":
      return False
    return True

  def get_data_shape(self, key):
    """
    :returns get_data(*, key).shape[1:], i.e. num-frames excluded
    :rtype: list[int]
    """
    if key in self.num_outputs:
      if self.num_outputs[key][1] <= 1:
        return []
      res_shape = [None] * (self.num_outputs[key][1] - 1)  # type: typing.List[typing.Union[None,int]]
      if not self.is_data_sparse(key):
        res_shape[-1] = self.get_data_dim(key)
      return res_shape
    if self.is_data_sparse(key):
      return []
    return [self.get_data_dim(key)]

  def have_seqs(self):
    """
    :return: whether num_seqs > 0
    :rtype: bool
    """
    return self.is_less_than_num_seqs(0)

  def len_info(self):
    """
    :rtype: str
    :returns a string to present the user as information about our len.
    Depending on our implementation, we can give some more or some less information.
    """
    return ", ".join([self.__class__.__name__,
                      "sequences: %s" % try_run(lambda: self.num_seqs, default="unknown"),
                      "frames: %s" % try_run(self.get_num_timesteps, default="unknown")])

  def is_less_than_num_seqs(self, n):
    """
    :type n: int
    :rtype: bool
    :returns whether n < num_seqs. In case num_seqs is not known in advance, it will wait
    until it knows that n is behind the end or that we have the seq.
    """
    # We keep this dynamic so that other implementations which don't know the num of seqs
    # in advance can handle this somehow.
    return n < self.num_seqs

  def can_serialize_data(self, key):
    """
    :param str key: e.g. "classes"
    :rtype: bool
    """
    return key in self.labels

  def serialize_data(self, key, data):
    """
    In case you have a :class:`Vocabulary`, just use :func:`Vocabulary.get_seq_labels`.

    :param str key: e.g. "classes". self.labels[key] should be set
    :param numpy.ndarray data: 0D or 1D
    :rtype: str
    """
    vocab = Vocabulary.create_vocab_from_labels(self.labels[key])
    if data.ndim == 0:
      return vocab.labels[data]
    return vocab.get_seq_labels(data)

  def calculate_priori(self, target="classes"):
    """
    :param str target:
    :rtype: numpy.ndarray
    """
    priori = numpy.zeros((self.num_outputs[target][0],), dtype=numpy.float32)
    i = 0
    while self.is_less_than_num_seqs(i):
      self.load_seqs(i, i + 1)
      for t in self.get_targets(target, i):
        priori[t] += 1
      i += 1
    return numpy.array(priori / self.get_num_timesteps(), dtype=numpy.float32)

  def iterate_seqs(self, recurrent_net=True, used_data_keys=None):
    """
    Takes chunking into consideration.

    :param bool recurrent_net: whether the order of frames matter
    :param set(str)|None used_data_keys:
    :return: generator which yields tuples (seq index, seq start, seq end)
    :rtype: list[(int,NumbersDict,NumbersDict)]
    """
    if self.custom_chunking_func:
      sentinel_kw = {"__fwd_compatible_random_arg_%i" % int(random() * 100): None}
      for seq_idx, t_start, t_end in self.custom_chunking_func(
            dataset=self, seq_idx_start=0, recurrent_net=recurrent_net, used_data_keys=used_data_keys, **sentinel_kw):
        yield seq_idx, t_start, t_end
      return

    chunk_size = self.chunk_size
    chunk_step = self.chunk_step
    if not recurrent_net:
      if chunk_size:
        print("Non-recurrent network, chunk size %s:%s ignored" % (chunk_size, chunk_step), file=log.v4)
        chunk_size = 0
    chunk_size = NumbersDict(chunk_size)
    chunk_step = NumbersDict(chunk_step)
    chunk_size_orig = chunk_size.copy()
    chunk_step_orig = chunk_step.copy()

    s = 0
    while self.is_less_than_num_seqs(s):
      length = self.get_seq_length(s)
      if chunk_size == 0:
        yield s, NumbersDict.constant_like(0, numbers_dict=length), length
      else:
        default_key = "data"
        if used_data_keys is not None:
          length = NumbersDict({k: length[k] for k in used_data_keys})
          if default_key not in used_data_keys:
            default_key = sorted(used_data_keys)[0]
          if chunk_step[default_key] == 0:  # allow some keys with zero chunk-step
            assert chunk_step.max_value() > 0
            default_key = [key for key in sorted(used_data_keys) if chunk_step[key] > 0][0]
          if self.chunking_variance > 0:
            chunking_variance = 1. - self.rnd_seq_drop.random() * self.chunking_variance
            for k in used_data_keys:
              chunk_size[k] = max(int(chunk_size_orig[k] * chunking_variance), 1)
              chunk_step[k] = max(int(chunk_step_orig[k] * chunking_variance), 1)
            # In case there are data keys with different chunk sizes,
            # make sure we keep the original ratio.
            smallest_key = [
              k for k in sorted(used_data_keys, key=lambda key: (chunk_step_orig[key], key))
              if chunk_step_orig[k] > 0][0]
            for k in used_data_keys:
              if chunk_size_orig[k] > chunk_size_orig[smallest_key]:
                if chunk_size_orig[k] % chunk_size_orig[smallest_key] == 0:
                  ratio = chunk_size_orig[k] // chunk_size_orig[smallest_key]
                  chunk_size[k] = chunk_size[smallest_key] * ratio
                  chunk_step[k] = chunk_step[smallest_key] * ratio
        assert chunk_step[default_key] > 0
        t = NumbersDict.constant_like(0, numbers_dict=length)
        # There are usually the 'data' (input) and 'classes' (targets) data-keys in `length` but there can be others.
        # We expect them all of the same length so that we can do chunking.
        # In case that some length is 0 or 1,
        # we treat it special and always return the full seq repeated for every chunk.
        keys_with_full_seqs = []
        for key in length.keys():
          if chunk_step[key] == chunk_step[default_key]:
            if length[key] == length[default_key]:
              continue  # ok
          if length[key] <= 1:  # special case as explained above
            keys_with_full_seqs.append(key)
            continue
          if chunk_step[key] == chunk_step[default_key]:
            raise Exception("Chunking with multiple data-keys of different length: %r" % length)
          else:
            limit = limit_default = 1
            if self.min_chunk_size == chunk_size[default_key]:
              limit = chunk_size[key]
              limit_default = chunk_size[default_key]
            nr_of_chunks = (length[key] - limit) // chunk_step[key] + 1
            nr_of_chunks_default = (length[default_key] - limit_default) // chunk_step[default_key] + 1
            assert nr_of_chunks == nr_of_chunks_default, (
              "%s: iterate seqs with chunking: length %r, chunk size/step %r/%r (min %r), key %r (default %r)" % (
                self, length, chunk_size, chunk_step, self.min_chunk_size, key, default_key))
        while length[default_key] > t[default_key]:
          chunk_start = NumbersDict(t)
          chunk_end = NumbersDict.min([t + chunk_size, length])
          for key in keys_with_full_seqs:
            chunk_start[key] = 0
            chunk_end[key] = length[key]
          if length.value is None:
            chunk_start.value = None
            chunk_end.value = None
          yield s, chunk_start, chunk_end
          t += chunk_step
          if length[default_key] - t[default_key] <= self.min_chunk_size:
            break
      s += 1

  def get_start_end_frames_full_seq(self, seq_idx):
    """
    :param int seq_idx:
    :return: (start,end) frame, taking context_window into account
    :rtype: (NumbersDict,NumbersDict)
    """
    end = self.get_seq_length(seq_idx)
    start = NumbersDict.constant_like(0, numbers_dict=end)
    start -= self.ctx_left
    end += self.ctx_right
    return start, end

  def sample(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: bool
    """
    if seq_idx in self.weights:
      weight = self.weights[seq_idx]
      return weight[0] >= weight[1]
    return True

  def update_weights(self, seqs, weights):
    """
    :param list[EngineBatch.BatchSeqCopyPart] seqs:
    :param list[float] weights:
    """
    for seq, weight in zip(seqs, weights):
      self.weights[seq.seq_idx] = [weight, 0]

  def _generate_batches(self, recurrent_net,
                        batch_size, max_seqs=-1, max_seq_length=sys.maxsize,
                        max_pad_size=None,
                        min_seq_length=0, pruning=0.0,
                        seq_drop=0.0, max_total_num_seqs=-1,
                        used_data_keys=None):
    """
    :param bool recurrent_net: If True, the batch might have a batch seq dimension > 1.
      Otherwise, the batch seq dimension is always 1 and multiple seqs will be concatenated.
    :param int|dict[str,int]|NumbersDict batch_size: Max number of frames in one batch.
    :param int|dict[str,int]|NumbersDict max_pad_size: Max number of zero-padded frames in one batch.
    :param int max_seqs: Max number of seqs per batch.
    :param int max_total_num_seqs:
    :param int|dict[str,int]|NumbersDict max_seq_length:
    :param set(str)|None used_data_keys:
    """
    if not batch_size:
      batch_size = sys.maxsize
    batch_size = NumbersDict(batch_size)
    assert not batch_size.any_compare(NumbersDict(0), (lambda a, b: a <= b))
    max_pad_size = NumbersDict(max_pad_size)
    if max_seqs == -1:
      max_seqs = float('inf')
    if not max_seq_length:
      max_seq_length = sys.maxsize
    if isinstance(max_seq_length, int) and max_seq_length < 0:
      max_seq_length = {"classes": -max_seq_length}
    max_seq_length = NumbersDict(max_seq_length)
    min_seq_length = NumbersDict(min_seq_length)
    assert max_seqs > 0
    assert seq_drop <= 1.0
    if not max_total_num_seqs or max_total_num_seqs < 0:
      max_total_num_seqs = float("inf")
    batch = Batch()
    total_num_seqs = 0
    last_seq_idx = -1
    avg_weight = sum([v[0] for v in self.weights.values()]) / (len(self.weights.keys()) or 1)
    for idx in self.weights:
      self.weights[idx][1] = random() * avg_weight * pruning
      self.weights[idx][0] *= (1. + pruning)
    for seq_idx, t_start, t_end in self.iterate_seqs(recurrent_net=recurrent_net, used_data_keys=used_data_keys):
      if not self.sample(seq_idx):
        continue
      if total_num_seqs > max_total_num_seqs:
        break
      t_start -= self.ctx_left
      t_end += self.ctx_right
      if recurrent_net:
        length = t_end - t_start
        if length.any_compare(max_seq_length, (lambda a, b: a > b)):
          continue
        if length.any_compare(min_seq_length, (lambda a, b: a < b)):
          continue
        if length.any_compare(batch_size, (lambda a, b: a > b)):
          print("warning: sequence length (%r) larger than limit (%r)" % (length, batch_size), file=log.v4)
        if self.rnd_seq_drop.random() < seq_drop:
          continue
        dt, ds = batch.try_sequence_as_slice(length)
        if batch.num_slices >= 1:
          if (dt * ds).any_compare(batch_size, (lambda a, b: a > b)):
            yield batch
            batch = Batch()
          elif ds > max_seqs:
            yield batch
            batch = Batch()
          elif (dt * ds - batch.get_total_num_frames() - length).any_compare(max_pad_size, (lambda a, b: a > b)):
            yield batch
            batch = Batch()
        batch.add_sequence_as_slice(seq_idx=seq_idx, seq_start_frame=t_start, length=length)
      else:  # Not recurrent.
        while t_start.max_value() < t_end.max_value():
          length = t_end - t_start
          num_frames = NumbersDict.min(
            [length, batch_size.copy_like(length) - batch.get_all_slices_num_frames().copy_like(length)])
          assert num_frames.max_value() > 0
          batch.add_frames(seq_idx=seq_idx, seq_start_frame=t_start, length=num_frames)
          if (
                batch.get_all_slices_num_frames().any_compare(batch_size, (lambda a, b: a >= b))
                or batch.get_num_seqs() > max_seqs):
            yield batch
            batch = Batch()
          t_start += num_frames
      if seq_idx != last_seq_idx:
        last_seq_idx = seq_idx
        total_num_seqs += 1

    if batch.get_all_slices_num_frames().max_value() > 0:
      yield batch

  def batch_set_generator_cache_whole_epoch(self):
    """
    The BatchSetGenerator can cache the list of batches which we generated across epochs.
    See self.generate_batches() and self._generate_batches().
    In many cases, the dataset does not support this, and in that case,
    it is not needed to enable this cache and waste memory.
    Caching it together with option shuffle_batches could also mean that
    there will be self.load_seqs() calls with non-monotonic seq-idxs.
    The only dataset currently which enables this is CachedDataset and thus HDFDataset.

    :return: whether we should enable this cache
    :rtype: bool
    """
    return False

  def generate_batches(self, shuffle_batches=False, **kwargs):
    """
    :param bool shuffle_batches:
    :param kwargs: will be passed to :func:`_generate_batches`
    :rtype: BatchSetGenerator
    """
    return BatchSetGenerator(
      dataset=self,
      generator=self._generate_batches(**kwargs),
      shuffle_batches=shuffle_batches,
      cache_whole_epoch=self.batch_set_generator_cache_whole_epoch())

  @classmethod
  def index_shape_for_batches(cls, batches, data_key="data"):
    """
    :param list[EngineBatch.Batch] batches:
    :param str data_key:
    :return: shape as (time, batch)
    :rtype: (int, int)
    """
    shape = [0, 0]  # time, batch
    for batch in batches:
      shape = [max(shape[0], batch.max_num_frames_per_slice[data_key]), shape[1] + batch.num_slices]
    return tuple(shape)


class DatasetSeq:
  """
  Encapsulates all data for one sequence.
  """

  def __init__(self, seq_idx, features, targets=None, ctc_targets=None, seq_tag=None):
    """
    :param int seq_idx: sorted seq idx in the Dataset
    :param numpy.ndarray|dict[str,numpy.ndarray] features: format 2d (time,feature) (float)
    :param dict[str,numpy.ndarray]|numpy.ndarray|None targets: name -> format 1d (time) (idx of output-feature)
    :param numpy.ndarray|None ctc_targets: format 1d (time) (idx of output-feature)
    :param str seq_tag: sequence name / tag
    """
    assert isinstance(seq_idx, int)
    self.seq_idx = seq_idx
    self.seq_tag = seq_tag or ("seq-%i" % seq_idx)
    if not isinstance(features, dict):
      assert isinstance(features, numpy.ndarray)
      features = {"data": features}
      if targets is None:
        targets = {}
      if isinstance(targets, numpy.ndarray):  # old format
        targets = {"classes": targets}
      assert isinstance(targets, dict)
      features.update(targets)
      targets = None
    assert isinstance(features, dict) and targets is None
    for v in features.values():
      assert isinstance(v, numpy.ndarray)
    self.features = features
    self.ctc_targets = ctc_targets

  @property
  def num_frames(self):
    """
    :rtype: NumbersDict
    """
    d = {k: (v.shape[0] if v.ndim >= 1 else 1)
         for (k, v) in self.features.items()}
    return NumbersDict(d)

  def get_data(self, key):
    """
    :param str key:
    :rtype: numpy.ndarray
    """
    return self.features[key]

  def get_data_keys(self):
    """
    :rtype: set[str]
    """
    return self.features.keys()

  def __repr__(self):
    return "<DataCache seq_idx=%i>" % self.seq_idx


def get_dataset_class(name):
  """
  :param str|type name:
  :rtype: type[Dataset]
  """
  if isinstance(name, type):
    assert issubclass(name, Dataset)
    return name

  from importlib import import_module
  # Only those modules which make sense to be loaded by the user,
  # because this function is only used for such cases.
  mod_names = [
    "hdf", "sprint", "generating", "numpy_dump",
    "meta", "lm", "stereo", "raw_wav", "map"]
  for mod_name in mod_names:
    mod = import_module("returnn.datasets.%s" % mod_name)
    if name in vars(mod):
      clazz = getattr(mod, name)
      assert issubclass(clazz, Dataset)
      return clazz
  return None


def init_dataset(kwargs, extra_kwargs=None, default_kwargs=None):
  """
  :param dict[str]|str|(()->dict[str])|Dataset kwargs:
  :param dict[str]|None extra_kwargs:
  :param dict[str]|None default_kwargs:
  :rtype: Dataset
  """
  assert kwargs
  if isinstance(kwargs, Dataset):
    data = kwargs
    data.initialize()
    return data
  if callable(kwargs):
    return init_dataset(kwargs(), extra_kwargs=extra_kwargs, default_kwargs=default_kwargs)
  if isinstance(kwargs, (str, unicode)):
    if kwargs.startswith("{"):
      kwargs = eval(kwargs)
    elif kwargs.startswith("config:"):
      from returnn.config import get_global_config
      config = get_global_config()
      data = eval(kwargs[len("config:"):], config.typed_dict, config.typed_dict)
      return init_dataset(data, extra_kwargs=extra_kwargs, default_kwargs=default_kwargs)
    else:
      config_str = kwargs
      kwargs = {}
      if default_kwargs:
        kwargs.update(default_kwargs)
      if extra_kwargs:
        kwargs.update(extra_kwargs)
      return init_dataset_via_str(config_str=config_str, **kwargs)
  assert isinstance(kwargs, dict)
  kwargs = kwargs.copy()
  assert "class" in kwargs
  clazz_name = kwargs.pop("class")
  clazz = get_dataset_class(clazz_name)
  if not clazz:
    raise Exception("Dataset class %r not found" % clazz_name)
  if default_kwargs:
    for key, value in default_kwargs.items():
      kwargs.setdefault(key, value)
  if extra_kwargs:
    kwargs.update(extra_kwargs)
  obj = clazz(**kwargs)
  assert isinstance(obj, Dataset)
  obj.initialize()
  return obj


def init_dataset_via_str(config_str, config=None, cache_byte_size=None, **kwargs):
  """
  :param str config_str: hdf-files, or "LmDataset:..." or so
  :param Config.Config|None config: optional, only for "sprint:..."
  :param int|None cache_byte_size: optional, only for HDFDataset
  :rtype: Dataset
  """
  kwargs = kwargs.copy()
  if 'window' not in kwargs and config and config.has('window'):
    kwargs['window'] = config.int('window', 1)
  from returnn.datasets.hdf import HDFDataset
  if config_str.startswith("sprint:"):
    kwargs["sprintConfigStr"] = config_str[len("sprint:"):]
    assert config, "need config for dataset in 'sprint:...' format. or use 'ExternSprintDataset:...' instead"
    sprint_trainer_exec_path = config.value("sprint_trainer_exec_path", None)
    assert sprint_trainer_exec_path, "specify sprint_trainer_exec_path in config"
    kwargs["sprintTrainerExecPath"] = sprint_trainer_exec_path
    from returnn.datasets.sprint import ExternSprintDataset
    cls = ExternSprintDataset
  elif config_str.startswith("config:"):
    from returnn.config import get_global_config
    if not config:
      config = get_global_config()
    data = eval(config_str[len("config:"):], config.typed_dict, config.typed_dict)
    return init_dataset(data, extra_kwargs=kwargs)
  elif ":" in config_str:
    kwargs.update(eval("dict(%s)" % config_str[config_str.find(":") + 1:]))
    class_name = config_str[:config_str.find(":")]
    cls = get_dataset_class(class_name)
  else:
    if cache_byte_size is not None:
      kwargs["cache_byte_size"] = cache_byte_size
    cls = HDFDataset
  if config:
    data = cls.from_config(config, **kwargs)
  else:
    data = cls(**kwargs)
  if isinstance(data, HDFDataset):
    for f in config_str.split(","):
      if f:
        assert os.path.exists(f)
        data.add_file(f)
  data.initialize()
  return data


def convert_data_dims(data_dims, leave_dict_as_is=False):
  """
  This converts what we called num_outputs originally,
  from the various formats which were allowed in the past
  (just an int, or dict[str,int]) into the format which we currently expect.
  In all cases, the output will be a new copy of the dict.

  :param int|dict[str,int|(int,int)|dict] data_dims: what we called num_outputs originally
  :param bool leave_dict_as_is:
  :rtype: dict[str,(int,int)|dict]
  :returns dict data-key -> (data-dimension, len(shape) (1 ==> sparse))
   (or potentially data-key -> dict, if leave_dict_as_is is True; for TensorFlow)
  """
  if isinstance(data_dims, int):
    data_dims = {"classes": data_dims}
  assert isinstance(data_dims, dict)
  data_dims = data_dims.copy()
  for k, v in list(data_dims.items()):
    if isinstance(v, int):
      v = (v, 2 if k == "data" else 1)
      data_dims[k] = v
    if isinstance(v, dict) and leave_dict_as_is:
      continue
    assert isinstance(v, (tuple, list))
    data_dims[k] = tuple(v)
    assert len(v) == 2
    assert isinstance(v[0], int)
    assert isinstance(v[1], int)
    assert 1 <= v[1]
  return data_dims


def shapes_for_batches(batches, data_keys, dataset=None, extern_data=None, enforce_min_len1=False):
  """
  :param list[EngineBatch.Batch] batches:
  :param list[str] data_keys:
  :param Dataset dataset:
  :param TFNetwork.ExternData extern_data: detailed data description. only used for TensorFlow
  :param bool enforce_min_len1:
  :rtype: dict[str,list[int]] | None
  """
  assert dataset or extern_data
  all_data_keys = set(data_keys)

  # The final device.data.shape is in format (time,batch,feature) in case of Theano.
  shape = [NumbersDict(0), 0]  # time,batch
  for batch in batches:
    shape = [NumbersDict.max([shape[0], batch.max_num_frames_per_slice]), shape[1] + batch.num_slices]
  if shape[1] == 0:
    return None
  assert shape[0].max_value() > 0
  # Theano has some buggy behaviour with tensors with some shape of zero.
  # We will just use one dummy frame in that case.
  # The index will stay zero in that case. (see EngineUtil.assign_dev_data())
  # However, also see the OutputLayer.output_index() behavior for forwarding.
  if not extern_data or enforce_min_len1:  # not needed if TensorFlow is used
    for k in all_data_keys:
      shape[0][k] = max(shape[0][k], 1)

  if extern_data:
    d = {}
    for k in all_data_keys:
      data_shape = list(extern_data.data[k].batch_shape)
      data_shape[extern_data.data[k].batch_dim_axis] = shape[1]
      if extern_data.data[k].have_time_axis():
        data_shape[extern_data.data[k].time_dim_axis] = shape[0][k]
      assert all([n is not None for n in data_shape]), "data %r" % extern_data.data[k]
      d[k] = data_shape
  else:  # shape via dataset
    d = {k: [shape[0][k], shape[1]] for k in all_data_keys}
    for k in all_data_keys:
      d[k] += dataset.get_data_shape(k)
  return d


def set_config_num_inputs_outputs_from_dataset(config, dataset):
  """
  :param Config.Config config:
  :param Dataset dataset:
  """
  from returnn.util import BackendEngine
  if BackendEngine.is_tensorflow_selected():
    # TF supports more fine-grained specification,
    # however the dataset does not store that in num_outputs.
    from returnn.tf.network import ExternData
    config.set("extern_data", {
      key: ExternData.data_kwargs_from_dataset_key(dataset=dataset, key=key)
      for key in dataset.get_data_keys()})
  else:
    config.set("num_inputs", dataset.num_inputs)
    config.set("num_outputs", dataset.num_outputs)




import theano
import theano.sandbox.cuda
import theano.tensor as T
from theano.compile import ViewOp
from theano import gof
from theano.compile import optdb
import numpy


def monkey_patches():
  """
  Applies a number of monkey patches,
  to make Theano work.
  """
  monkey_patch_numpy_get_ndarray_c_version()


def monkey_patch_numpy_get_ndarray_c_version():
  """
  Error:

    AttributeError: module 'numpy.core.multiarray' has no attribute '_get_ndarray_c_version'

  In Theano, when np.core.multiarray._get_ndarray_c_version() is called.

  Link:
    https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=918090
  """
  import numpy.core.multiarray as ma
  if not hasattr(ma, "_get_ndarray_c_version"):  # later than 1.16.0
    print("Monkey patch Numpy _get_ndarray_c_version.")
    # Instead of patching Theano, we patch Numpy, which is simpler.
    import numpy.core._multiarray_umath as mau
    ma._get_ndarray_c_version = mau._get_ndarray_c_version


def time_batch_make_flat(val):
  """
  :rtype val: theano.Variable
  :rtype: theano.Variable

  Will flatten the first two dimensions and leave the others as is.
  """
  assert val.ndim > 1
  s0 = val.shape[0] * val.shape[1]
  newshape = [s0] + [val.shape[i] for i in range(2, val.ndim)]
  return T.reshape(val,
                   newshape,
                   ndim=val.ndim - 1)
                   #name="flat_%s" % val.name)


def class_idx_seq_to_1_of_k(seq, num_classes, dtype="float32"):
  """
  :param theano.Variable seq: ndarray with indices
  :param int | theano.Variable num_classes: number of classes
  :param str dtype: eg "float32"
  :rtype: theano.Variable
  :returns ndarray with one added dimension of size num_classes.
  That is the one-hot-encoding.
  This function is like theano.tensor.extra_ops.to_one_hot
  but we can handle multiple dimensions.
  """
  shape = [seq.shape[i] for i in range(seq.ndim)] + [num_classes]
  eye = T.eye(num_classes, dtype=dtype)
  m = eye[T.cast(seq, 'int32')].reshape(shape)
  return m


def tiled_eye(n1, n2, dtype="float32"):
  r1 = T.maximum((n1 - 1) / n2 + 1, 1)
  r2 = T.maximum((n2 - 1) / n1 + 1, 1)
  small_eye = T.eye(T.minimum(n1, n2), dtype=dtype)
  tiled_big = T.tile(small_eye, (r1, r2))
  tiled_part = tiled_big[:n1,:n2]
  return tiled_part


def opt_contiguous_on_gpu(x):
  if theano.sandbox.cuda.cuda_enabled:
    return theano.sandbox.cuda.basic_ops.gpu_contiguous(x)
  return x


def windowed_batch(source, window):
  """
  :param theano.TensorVariable source: 3d tensor of shape (n_time, n_batch, n_dim)
  :param int|theano.Variable window: window size
  :return: tensor of shape (n_time, n_batch, window * n_dim)
  """
  assert source.ndim == 3  # (time,batch,dim). not sure how to handle other cases
  n_time = source.shape[0]
  n_batch = source.shape[1]
  n_dim = source.shape[2]
  w_right = window // 2
  w_left = window - w_right - 1
  pad_left = T.zeros((w_left, n_batch, n_dim), dtype=source.dtype)
  pad_right = T.zeros((w_right, n_batch, n_dim), dtype=source.dtype)
  padded = T.concatenate([pad_left, source, pad_right], axis=0)  # shape[0] == n_time + window - 1
  tiled = T.tile(padded, (1, 1, window))  # shape[2] == n_dim * window
  tiled_reshape = T.reshape(tiled, ((n_time + window - 1), n_batch, window, n_dim))
  # We want to shift every dim*time block by one to the left.
  # To do this, we interpret that we have one more time frame (i.e. n_time+window).
  # We have to do some dimshuffling so that we get the right layout, then we can flatten,
  # add some padding, and then dimshuffle it back.
  # Then we can take out the first n_time frames.
  tiled_dimshuffle = tiled_reshape.dimshuffle(2, 0, 1, 3)  # (window,n_time+window-1,batch,dim)
  tiled_flat = T.flatten(tiled_dimshuffle)
  rem = n_batch * n_dim * window
  tiled_flat_pad_right = T.concatenate([tiled_flat, T.zeros((rem,), dtype=source.dtype)])
  tiled_reshape_shift = T.reshape(tiled_flat_pad_right, (window, n_time + window, n_batch, n_dim))  # add time frame
  final_dimshuffle = tiled_reshape_shift.dimshuffle(1, 2, 0, 3)  # (n_time+window,batch,window,dim)
  final_sub = final_dimshuffle[:n_time]  # (n_time,batch,window,dim)
  final_concat_dim = final_sub.reshape((n_time, n_batch, window * n_dim))
  return final_concat_dim


def delta_batch(source, window):
  """
  :param theano.TensorVariable source: 3d tensor of shape (n_time, n_batch, n_dim)
  :param int|theano.Variable window: window size
  :return: tensor of shape (n_time, n_batch, window * n_dim)
  Similar to numpy.diff. Also called delta.
  TODO with conv op
  """
  assert source.ndim == 3  # (time,batch,dim). not sure how to handle other cases
  w = windowed_batch(source, window + 1)
  return w[1:] - w[:-1]


def context_batched(source, window):
  """
  same as windowed_batch but with window center at the end of the window
  :param theano.TensorVariable source: 3d tensor of shape (n_time, n_batch, n_dim)
  :param int|theano.Variable window: window size
  :return: tensor of shape (n_time, n_batch, window * n_dim)
  """
  assert source.ndim == 3  # (time,batch,dim). not sure how to handle other cases
  n_time = source.shape[0]
  n_batch = source.shape[1]
  n_dim = source.shape[2]
  w_left = window - 1
  pad_left = T.zeros((w_left, n_batch, n_dim), dtype=source.dtype)
  padded = T.concatenate([pad_left, source], axis=0)  # shape[0] == n_time + window - 1
  tiled = T.tile(padded, (1, 1, window))  # shape[2] == n_dim * window
  tiled_reshape = T.reshape(tiled, ((n_time + window - 1), n_batch, window, n_dim))
  # We want to shift every dim*time block by one to the left.
  # To do this, we interpret that we have one more time frame (i.e. n_time+window).
  # We have to do some dimshuffling so that we get the right layout, then we can flatten,
  # add some padding, and then dimshuffle it back.
  # Then we can take out the first n_time frames.
  tiled_dimshuffle = tiled_reshape.dimshuffle(2, 0, 1, 3)  # (window,n_time+window-1,batch,dim)
  tiled_flat = T.flatten(tiled_dimshuffle)
  rem = n_batch * n_dim * window
  tiled_flat_pad_right = T.concatenate([tiled_flat, T.zeros((rem,), dtype=source.dtype)])
  tiled_reshape_shift = T.reshape(tiled_flat_pad_right, (window, n_time + window, n_batch, n_dim))  # add time frame
  final_dimshuffle = tiled_reshape_shift.dimshuffle(1, 2, 0, 3)  # (n_time+window,batch,window,dim)
  final_sub = final_dimshuffle[:n_time]  # (n_time,batch,window,dim)
  final_concat_dim = final_sub.reshape((n_time, n_batch, window * n_dim))
  return final_concat_dim

def window_batch_timewise(t,b,w,full_index):
  for i in range(w):
    full_index = T.set_subtensor(full_index[i], T.roll(full_index[i], i))
    if i > 0:
      full_index = T.inc_subtensor(full_index[i], T.where(full_index[i] > 0, i * t * b - i, 0))
  return full_index

def slice_for_axis(axis, s):
  return (slice(None),) * axis + (s,)


def downsample(source, axis, factor, method="average"):
  assert factor == int(factor), "factor is expected to be an int"
  factor = int(factor)
  # make shape[axis] a multiple of factor
  src = source
  source = source[slice_for_axis(axis=axis, s=slice(0, (source.shape[axis] // factor) * factor))]
  # Add a temporary dimension as the factor.
  added_dim_shape = [source.shape[i] for i in range(source.ndim)]
  added_dim_shape = added_dim_shape[:axis] + [source.shape[axis] // factor, factor] + added_dim_shape[axis + 1:]
  if method == "lstm":
    assert axis == 0
    return source
  source = T.reshape(source, added_dim_shape)
  if method == "average":
    return T.mean(source, axis=axis + 1)
  elif method == "max":
    return T.max(source, axis=axis + 1)
  elif method == "min":
    return T.min(source, axis=axis + 1)
  elif method == "concat" or method == 'mlp': # concatenates in last dimension
    #return T.reshape(source, added_dim_shape[:axis+1] + added_dim_shape[axis+2:-1] + [added_dim_shape[-1] * factor])
    return source.swapaxes(axis+1,src.ndim-1).reshape([source.shape[0],src.shape[1]] + [factor * source.shape[3]])
  elif method == "batch":
    assert axis == 0
    return source.dimshuffle(1,0,2,3).reshape((source.shape[1],source.shape[0]*source.shape[2],source.shape[3]))
  else:
    assert False, "unknown downsample method %r" % method


def upsample(source, axis, factor, method="nearest-neighbor", target_axis_len=None):
  if method == "nearest-neighbor":
    assert factor == int(factor), "factor is expected to be an int. not implemented otherwise yet."
    factor = int(factor)
    target = T.repeat(source, factor, axis=axis)
    if target_axis_len is not None:
      # We expect that we need to add a few frames. Just use the last frame.
      last = source[slice_for_axis(axis=axis, s=slice(-1, None))]
      target = pad(target, axis=axis, target_axis_len=target_axis_len, pad_value=last)
    return target
  else:
    assert False, "unknown upsample method %r" % method


def pad(source, axis, target_axis_len, pad_value=None):
  if pad_value is None:
    pad_value = T.zeros([source.shape[i] if i != axis else 1 for i in range(source.ndim)], dtype=source.dtype)
  target_axis_len = T.cast(target_axis_len, dtype="int32")
  num_missing = target_axis_len - source.shape[axis]
  # There is some strange bug in Theano. If num_missing is 0, in some circumstances,
  # it crashes with Floating point exception.
  # Thus, do this workaround.
  num_missing = T.maximum(num_missing, 1)
  target = T.concatenate([source, T.repeat(pad_value, num_missing, axis=axis)], axis=axis)
  # Because of the workaround, we need this.
  target = target[slice_for_axis(axis=axis, s=slice(0, target_axis_len))]
  return target


def chunked_time_reverse(source, chunk_size):
  """
  :param source: >=1d array (time,...)
  :param chunk_size: int
  :return: like source
  Will not reverse the whole time-dim, but only every time-chunk.
  E.g. source=[0 1 2 3 4 5 6], chunk_size=3, returns [2 1 0 5 4 3 0].
  (Padded with 0, recovers original size.)
  """
  chunk_size = T.cast(chunk_size, dtype="int32")
  num_chunks = (source.shape[0] + chunk_size - 1) // chunk_size
  needed_time = num_chunks * chunk_size
  remaining_dims = [source.shape[i + 1] for i in range(source.ndim - 1)]
  padded_source = pad(source, axis=0, target_axis_len=needed_time)
  reshaped = padded_source.reshape([num_chunks, chunk_size] + remaining_dims)
  reshaped_rev = reshaped[:, ::-1]
  rev_correct_ndim = reshaped_rev.reshape([needed_time] + remaining_dims)
  return rev_correct_ndim[:source.shape[0]]


def try_register_canonicalize(f):
  try:
    return T.opt.register_canonicalize(f)
  except ValueError as e:
    print("try_register_canonicalize warning: %s" % e)
    return f  # just ignore


def try_register_gpu_opt(base_op_class):
  def do_register_gpu_opt(f):
    from theano.sandbox.cuda import cuda_available, register_opt
    from theano.gof.opt import local_optimizer
    f = local_optimizer([base_op_class])(f)
    if cuda_available:
      f = register_opt()(f)
    return f  # just ignore
  return do_register_gpu_opt


class GradDiscardOutOfBound(ViewOp):
  # See also theano.gradient.GradClip for a similar Op.
  __props__ = ()
  def __init__(self, lower_bound, upper_bound):
    super(GradDiscardOutOfBound, self).__init__()
    # We do not put those member in __eq__ or __hash__
    # as they do not influence the perform of this op.
    self.lower_bound = lower_bound
    self.upper_bound = upper_bound
    assert(self.lower_bound <= self.upper_bound)

  def grad(self, args, g_outs):
    return [T.switch(T.or_(T.lt(g_out, self.lower_bound), T.gt(g_out, self.upper_bound)),
                     T.cast(0, dtype=g_out.dtype),
                     g_out)
            for g_out in g_outs]

def grad_discard_out_of_bound(x, lower_bound, upper_bound):
  return GradDiscardOutOfBound(lower_bound, upper_bound)(x)

@try_register_canonicalize
@theano.gof.local_optimizer([GradDiscardOutOfBound])
def _local_grad_discard(node):
  if isinstance(node.op, GradDiscardOutOfBound):
    return node.inputs



def gaussian_filter_1d(x, sigma, axis, window_radius=40):
  """
  Filter 1d input with a Gaussian using mode `nearest`.
  x is expected to be 2D/3D of type (time,batch,...).
  Adapted via: https://github.com/Theano/Theano/issues/3793
  Original Author: https://github.com/matthias-k
  """
  assert 2 <= x.ndim <= 3
  assert 0 <= axis < x.ndim

  # Construction of 1d kernel
  filter_1d = T.arange(2*window_radius + 1) - window_radius
  filter_1d = T.exp(-0.5*filter_1d**2/sigma**2)
  filter_1d = filter_1d / filter_1d.sum()
  filter_1d = filter_1d.astype(x.dtype)  # 1D, window-dim

  # transform `x` so that it is valid input for T.nnet.conv2d (batch size, stack size, nb row, nb col).
  # first, axis 1 of `x` because that is the number of batches.
  # then, a dummy stack size dim (not exactly sure about that. GpuDnnConv images and kernel must have the same stack size)
  # at the end the axis where we want to apply the conv on.
  # in between all remaining axes.
  # e.g., if axis==0 and x.ndim==3, we should get [1, 'x', 2, 0]
  blur_dims = [1] + ['x'] + [i for i in range(x.ndim) if i not in (1, axis)] + [axis]
  while len(blur_dims) < 4:
    blur_dims.insert(len(blur_dims) - 1, 'x')
  assert len(blur_dims) == 4
  blur_input = x.dimshuffle(blur_dims)
  assert blur_input.ndim == 4
  filter_W = filter_1d.dimshuffle(['x','x','x',0])

  # Construction of filter pipeline
  blur_input_start = blur_input[:, :, :, :1]
  blur_input_start_padding = T.repeat(blur_input_start, window_radius, axis=3)
  blur_input_end = blur_input[:, :, :, -1:]
  blur_input_end_padding = T.repeat(blur_input_end, window_radius, axis=3)

  padded_input = T.concatenate([blur_input_start_padding, blur_input, blur_input_end_padding], axis=3)

  # padded_input supposed to be 4D (batch size, stack size, nb row, nb col).
  # filter_W supposed to be 4D (nb filters, stack size, nb row, nb col).
  blur_op = T.nnet.conv2d(padded_input, filter_W, border_mode='valid',
                          filter_shape=[1, 1, 1, (2*window_radius + 1) if isinstance(window_radius, int) else None])
  # blur_op is 4D (batch size, nb filters, output row, output col).
  # output row = stack size * nb row.
  blur_op = blur_op[:, 0, :, :]  # only one filter, remove dimension
  # blur_op is 3D (batch size, output row, output col).
  y = blur_op.dimshuffle({0:2,2:1}[axis], 0, {0:1,2:2}[axis])
  if x.ndim == 2: y = y[:, :, 0]
  assert x.ndim == y.ndim
  return y


def log_sum_exp(x, axis):
  x_max = T.max(x, axis=axis)
  x_max_bc = T.makeKeepDims(x, x_max, axis=axis)
  return T.log(T.sum(T.exp(x - x_max_bc), axis=axis)) + x_max

def max_filtered(x, axis, index):
  index = T.cast(index, dtype="float32")  # 2D, time*batch
  index_bc = index.dimshuffle(*(range(index.ndim) + ['x'] * (x.ndim - index.ndim)))
  x_min = T.min(x, axis=axis, keepdims=True)
  x_filtered = x * index_bc + x_min * (numpy.float32(1) - index_bc)
  assert x_filtered.ndim == x.ndim
  x_max = T.max(x_filtered, axis=axis)  # we ignore the out-of-index frames
  return x_max

def log_sum_exp_index(x, axis, index):
  index = T.cast(index, dtype="float32")  # 2D, time*batch
  index_bc = index.dimshuffle(*(range(index.ndim) + ['x'] * (x.ndim - index.ndim)))
  assert index_bc.ndim == x.ndim
  x_max = max_filtered(x, axis=axis, index=index)  # we ignore the out-of-index frames
  x_max_bc = T.makeKeepDims(x, x_max, axis=axis)
  assert x.ndim == x_max_bc.ndim
  x_shift = (x - x_max_bc) * index_bc  # filter out out-of-index. exp() could be inf otherwise
  return T.log(T.sum(T.exp(x_shift) * index_bc, axis=axis)) + x_max


def global_softmax(z, index, mode):
  """
  :param theano.Variable z: 3D array. time*batch*feature
  :param theano.Variable index: 2D array, 0 or 1, time*batch
  :rtype: theano.Variable
  :returns 3D array. exp(z) / Z, where Z = sum(exp(z),axis=[0,2]) / z.shape[0].
  """
  assert z.ndim == 3
  assert index.ndim == 2
  index = T.cast(index, dtype="float32")  # 2D, time*batch
  index_bc = index.dimshuffle(0, 1, 'x')
  times = T.sum(index, axis=0)  # 1D, batch
  assert times.ndim == 1
  z_max2 = T.max(z, axis=2)
  z_max2_bc = z_max2.dimshuffle(0, 1, 'x')
  ez = T.exp(z - z_max2_bc)
  Z_frame = T.sum(ez, axis=2)  # 2D, time*batch
  if mode == "local":  # this is classic framewise softmax
    Z_log_norm_bc = T.log(Z_frame).dimshuffle(0, 1, 'x') + z_max2_bc
  elif mode == "log-norm":
    Z_log_frame = T.log(Z_frame) + z_max2
    Z_log_norm = T.sum(Z_log_frame * index, axis=0) / times  # log-normalized. 1D, batch
    Z_log_norm_bc = Z_log_norm.dimshuffle('x', 0, 'x')  # 3D, time*batch*feature
  elif mode == "maxshift-log-norm":
    Z_log_frame = T.log(Z_frame)
    Z_log_norm = T.sum(Z_log_frame * index, axis=0) / times  # log-normalized. 1D, batch
    Z_log_norm_bc = Z_log_norm.dimshuffle('x', 0, 'x')  # 3D, time*batch*feature
    Z_log_norm_bc = Z_log_norm_bc + z_max2_bc
  elif mode == "std-norm":
    #Z_log_norm = T.log( T.sum(T.exp(T.log(Z_frame) + z_max), axis=0) / times )  <- we want that
    Z_log_frame = T.log(Z_frame) + z_max2
    assert Z_log_frame.ndim == 2
    Z_log_norm = log_sum_exp_index(Z_log_frame, index=index, axis=0) - T.log(times)
    assert Z_log_norm.ndim == 1
    Z_log_norm_bc = Z_log_norm.dimshuffle('x', 0, 'x')  # 3D, time*batch*feature
  elif mode == "maxshift-std-norm":
    # We normalize each shifted frame.
    Z_norm = T.sum(Z_frame * index, axis=0) / times  # 1D, batch
    assert Z_norm.ndim == 1
    Z_log_norm = T.log(Z_norm)
    Z_log_norm_bc = Z_log_norm.dimshuffle('x', 0, 'x')  # 3D, time*batch*feature
    Z_log_norm_bc = Z_log_norm_bc + z_max2_bc
  elif mode.startswith("gauss-maxshift("):
    modeend = mode.find(")-")
    assert modeend >= 0
    sigma = float(mode[len("gauss-maxshift("):modeend])
    z_gmax2 = gaussian_filter_1d(z_max2, sigma=sigma, axis=0)
    z_gmax2_bc = z_gmax2.dimshuffle(0, 1, 'x')
    z = z - z_gmax2_bc
    return global_softmax(z - z_gmax2_bc, mode=mode[modeend + 2:], index=index)
  elif mode.startswith("gauss-std-norm("):
    modeend = mode.find(")")
    assert modeend >= 0 and modeend == len(mode) - 1
    sigma = float(mode[len("gauss-std-norm("):modeend])
    Z_log_frame = T.log(Z_frame) + z_max2
    assert Z_log_frame.ndim == 2
    Z_log_frame_g = gaussian_filter_1d(z_max2, sigma=sigma, axis=0)
    Z_log_norm = log_sum_exp_index(Z_log_frame_g, index=index, axis=0) - T.log(times)
    assert Z_log_norm.ndim == 1
    Z_log_norm_bc = Z_log_norm.dimshuffle('x', 0, 'x')  # 3D, time*batch*feature
  else:
    assert False, "invalid global_softmax mode %r" % mode
  return T.exp(z - Z_log_norm_bc)


def show_global_softmax_stats(z):
  """
  :param z: numpy.ndarray or Theano Var (eval-able), 2D time*features
  """
  def stats(y): return numpy.min(y), numpy.max(y), numpy.mean(y), numpy.var(y)
  if z.ndim == 3: z = z[:,0,:]
  assert z.ndim == 2
  z_numpy = z
  z = T.as_tensor_variable(z)
  if not isinstance(z, numpy.ndarray): z = z.eval()
  print("show_global_softmax_stats for shape %s" % (z_numpy.shape,))
  print(" z min/max/mean/var = %s" % (stats(z_numpy),))
  z_max1 = numpy.max(z_numpy, axis=1)
  print(" z max1 min/max/mean/var = %s" % (stats(z_max1),))
  z_dmax1 = z_max1[:-1] - z_max1[1:]
  print(" z dmax1 min/max/mean/var = %s" % (stats(z_dmax1),))
  z_gmax1 = gaussian_filter_1d(T.as_tensor_variable(z_max1).dimshuffle(0, 'x'), sigma=10.0, axis=0).eval()[:,0]
  print(" z gmax1 min/max/mean/var = %s" % (stats(z_gmax1),))
  z_dgmax1 = z_gmax1[:-1] - z_gmax1[1:]
  print(" z dgmax1 min/max/mean/var = %s" % (stats(z_dgmax1),))
  z = T.cast(z, "float32")  # we always expect this precision
  z = z.dimshuffle(0, 'x', 1)  # add batch-dim
  index = T.ones((z.shape[0], 1))
  for mode in ["local", "log-norm", "maxshift-log-norm", "std-norm", "maxshift-std-norm",
               "gauss-maxshift(2.0)-std-norm", "gauss-maxshift(5.0)-std-norm", "gauss-maxshift(2.0)-log-norm",
               "gauss-std-norm(2.0)", "gauss-std-norm(5.0)"]:
    print(" mode %s" % mode)
    y = global_softmax(z, index=index, mode=mode).eval()
    assert y.ndim == 3 and y.shape[1] == 1
    y = y[:,0,:]
    print("  min/max/mean/var = %s" % (stats(y),))
    log_y = numpy.log(y)
    print("  log min/max/mean/var = %s" % (stats(log_y),))
    y_sum1 = numpy.sum(y, axis=1)
    print("  sum1 min/max/mean/var = %s" % (stats(y_sum1),))
    log_y_sum1 = numpy.log(y_sum1)
    print("  log sum1 min/max/mean/var = %s" % (stats(log_y_sum1),))


def complex_elemwise_mult(a, b, axis=-1):
  assert a.ndim == b.ndim
  if axis < 0: axis %= a.ndim
  assert 0 <= axis < a.ndim
  a_real = a[slice_for_axis(axis=axis, s=slice(0, a.shape[axis] / 2))]
  a_imag = a[slice_for_axis(axis=axis, s=slice(a.shape[axis] / 2, None))]
  b_real = b[slice_for_axis(axis=axis, s=slice(0, b.shape[axis] / 2))]
  b_imag = b[slice_for_axis(axis=axis, s=slice(b.shape[axis] / 2, None))]
  r_real = a_real * b_real - a_imag * b_imag
  r_imag = a_real * b_imag + a_imag * b_real
  return T.concatenate([r_real, r_imag], axis=axis)

def complex_bound(a, axis=-1):
  # Via Associative LSTMs, https://arxiv.org/abs/1602.03032
  if axis < 0: axis %= a.ndim
  assert 0 <= axis < a.ndim
  a_real = a[slice_for_axis(axis=axis, s=slice(0, a.shape[axis] / 2))]
  a_imag = a[slice_for_axis(axis=axis, s=slice(a.shape[axis] / 2, None))]
  d = T.maximum(numpy.float32(1), T.sqrt(a_real * a_real + a_imag * a_imag))
  r_real = a_real / d
  r_imag = a_imag / d
  return T.concatenate([r_real, r_imag], axis=axis)

def complex_dot(a, b):
  assert a.ndim >= 1
  assert b.ndim >= 1
  a_axis = a.ndim - 1
  a_real = a[slice_for_axis(axis=a_axis, s=slice(0, a.shape[a_axis] / 2))]
  a_imag = a[slice_for_axis(axis=a_axis, s=slice(a.shape[a_axis] / 2, None))]
  b_axis = 0
  b_real = b[slice_for_axis(axis=b_axis, s=slice(0, b.shape[b_axis] / 2))]
  b_imag = b[slice_for_axis(axis=b_axis, s=slice(b.shape[b_axis] / 2, None))]
  r_real = T.dot(a_real, b_real) - T.dot(a_imag, b_imag)
  r_imag = T.dot(a_real, b_imag) + T.dot(a_imag, b_real)
  r_axis = r_real.ndim - 1
  return T.concatenate([r_real, r_imag], axis=r_axis)


def indices_in_flatten_array(ndim, shape, *args):
  """
  We expect that all args can be broadcasted together.
  So, if we have some array A with ndim&shape as given,
  A[args] would give us a subtensor.
  We return the indices so that A[args].flatten()
  and A.flatten()[indices] are the same.
  """
  assert ndim > 0
  assert len(args) == ndim
  indices_per_axis = [args[i] for i in range(ndim)]
  for i in range(ndim):
    for j in range(i + 1, ndim):
      indices_per_axis[i] *= shape[j]
  indices = indices_per_axis[0]
  for i in range(1, ndim):
    indices += indices_per_axis[i]
  return indices

def circular_convolution(a, b):
  from returnn.theano.device import is_using_gpu
  has_gpuarray = is_using_gpu()
  try:
    import pygpu
  except Exception:
    has_gpuarray = False
  if has_gpuarray:
    from theano.gpuarray.fft import curfft as fft
    from theano.gpuarray.fft import cuirfft as ifft
  else:
    from theano.tensor.fft import rfft as fft
    from theano.tensor.fft import irfft as ifft
  return ifft(fft(a) * fft(b))


def unroll_scan(fn, sequences=(), outputs_info=(), non_sequences=(), n_steps=None,
                go_backwards=False):
  """
  Helper function to unroll for loops. Can be used to unroll theano.scan.
  The parameter names are identical to theano.scan, please refer to here
  for more information.

  Note that this function does not support the truncate_gradient
  setting from theano.scan.

  Code adapted from https://github.com/Lasagne/Lasagne.
  Thank you!

  Parameters
  ----------

  fn : function
      Function that defines calculations at each step.

  sequences : TensorVariable or list of TensorVariables
      List of TensorVariable with sequence data. The function iterates
      over the first dimension of each TensorVariable.

  outputs_info : list of TensorVariables
      List of tensors specifying the initial values for each recurrent
      value.

  non_sequences: list of TensorVariables
      List of theano.shared variables that are used in the step function.

  n_steps: int
      Number of steps to unroll.

  go_backwards: bool
      If true the recursion starts at sequences[-1] and iterates
      backwards.

  Returns
  -------
  Tuple of the form (outputs, updates).
  outputs is a list of TensorVariables. Each element in the list gives the recurrent
  values at each time step.
  updates is an empty dict for now.
  """
  if not isinstance(sequences, (list, tuple)):
    sequences = [sequences]
  sequences = list(sequences)
  outputs_info = list(outputs_info)
  non_sequences = list(non_sequences)

  # When backwards reverse the recursion direction
  counter = range(n_steps)
  if go_backwards:
    counter = counter[::-1]

  output = []
  prev_vals = outputs_info
  until = []
  for i in counter:
    assert len(prev_vals) == len(outputs_info)
    prev_vals = [prev for prev, out_info in zip(prev_vals, outputs_info) if out_info is not None]
    step_input = [s[i] for s in sequences] + prev_vals + non_sequences
    out_ = fn(*step_input)
    # The returned values from step can be either a TensorVariable,
    # a list, or a tuple.  Below, we force it to always be a list.
    if isinstance(out_, T.TensorVariable):
      out_ = [out_]
    if isinstance(out_, tuple):
      if len(out_) >= 1 and isinstance(out_[0], (list, tuple)):
        if len(out_) >= 2:
          assert not out_[1], "shared var updates not supported"
        if len(out_) >= 3:
          assert isinstance(out_[2], theano.scan_module.until)
          until.append(T.neq(out_[2].condition, 0))
        out_ = list(out_[0])
      else:
        out_ = list(out_)
    output.append(out_)

    prev_vals = output[-1]

  # iterate over each scan output and convert it to same format as scan:
  # [[output11, output12,...output1n],
  # [output21, output22,...output2n],...]
  output_scan = []
  for i in range(len(output[0])):
    l = map(lambda x: x[i], output)
    output_scan.append(T.stack(*l))

  if until:
    assert len(until) == n_steps
    until_conds = T.stack(*until)
    new_len = T.switch(T.any(until_conds),
                       T.minimum(T.argmax(until_conds) + 1, n_steps),
                       n_steps)
    output_scan = [out[:new_len] for out in output_scan]

  if len(output_scan) == 1:
    output_scan = output_scan[0]
  updates = {}
  return output_scan, updates


def make_var_tuple(v):
  if isinstance(v, theano.Variable):
    return (v, )
  if isinstance(v, list):
    return tuple(v)
  assert isinstance(v, tuple)
  return v


class Contiguous(T.extra_ops.CpuContiguous):
  # Original CpuContiguous is missing the grad.
  # See https://github.com/Theano/Theano/issues/4399.
  def grad(self, inputs, output_grads):
    dout, = output_grads
    dout = T.as_tensor_variable(dout)
    return [dout]


@gof.local_optimizer([Contiguous], inplace=True)
def opt_remove_contiguous(node):
  if isinstance(node.op, Contiguous):
    x, = node.inputs
    if x.owner and isinstance(x.owner.op, (T.Alloc, T.AllocEmpty, T.extra_ops.CpuContiguous)):
      return [x]
  return False

optdb.register('opt_remove_contiguous',
               gof.TopoOptimizer(opt_remove_contiguous),
               10, 'fast_run')

# Theano will not do this optimization. So we register it now.
# See: https://github.com/Theano/Theano/issues/4400
@try_register_gpu_opt(Contiguous)
def local_gpu_Contiguous(node):
  if isinstance(node.op, Contiguous):
    # see also: https://github.com/Theano/Theano/blob/master/theano/sandbox/cuda/opt.py
    from theano.sandbox.cuda import host_from_gpu
    x, = node.inputs
    if x.owner and x.owner.op == host_from_gpu:
      from theano.sandbox.cuda.basic_ops import gpu_contiguous
      return [host_from_gpu(gpu_contiguous(x.owner.inputs[0]))]


class DumpOp(theano.Op):
  __props__ = ("filename", "with_grad")
  view_map = {0: [0]}

  def __init__(self, filename, container=None, with_grad=True, parent=None, step=1):
    super(DumpOp, self).__init__()
    self.filename = filename
    self.container = container
    self.with_grad = with_grad
    self.counter = 0
    self.parent = parent
    self.step = step

  def make_node(self, x):
    x = T.as_tensor_variable(x)
    return gof.Apply(self, [x], [x.type()])

  def perform(self, node, inputs, output_storage):
    x, = inputs
    self.dump(x)
    output_storage[0][0] = x

  def grad(self, inputs, output_grads):
    dout, = output_grads
    dout = T.as_tensor_variable(dout)
    if self.with_grad:
      # Note: This assumes that there will be only one such gradient.
      dout = DumpOp(filename=self.filename + ".grad", container=self.container, parent=self, step=self.step)(dout)
    return [dout]

  def dump(self, x):
    filename = self.get_full_filename()
    if filename is not None:
      if self.container is not None:
        assert filename not in self.container
        self.container[filename] = x.copy()
      else:
        import os, numpy
        assert not os.path.exists(filename), "%s already exists, not overwriting" % filename
        numpy.save(filename, x)

  def get_full_filename(self):
    counter = self.get_counter()
    self.inc_counter()
    if counter % self.step == 0:
      return "%s.%i.npy" % (self.filename, counter)
    else:
      return None

  def get_counter(self):
    if self.parent: return self.parent.get_counter() - 1
    return self.counter

  def inc_counter(self):
    if self.parent: return
    self.counter += 1


def softmax(z):
  dims = [z.shape[i] for i in range(z.ndim)]
  collapsed_dim = numpy.prod(dims[:-1])
  return T.nnet.softmax(z.reshape((collapsed_dim, dims[-1]))).reshape(dims)


def layer_normalization(x, bias=None, scale=None, eps=1e-5):
  """
  Layer Normalization, https://arxiv.org/abs/1607.06450
  x is mean and variance normalized along its feature dimension.
  After that, we allow a bias and a rescale. This is supposed to be trainable.
  :param x: 3d tensor (time,batch,dim) (or any ndim, last dim is expected to be dim)
  :param bias: 1d tensor (dim) or None
  :param scale: 1d tensor (dim) or None
  """
  mean = T.mean(x, axis=x.ndim - 1, keepdims=True)
  std = T.sqrt(T.var(x, axis=x.ndim - 1, keepdims=True) + numpy.float32(eps))
  assert mean.ndim == std.ndim == x.ndim
  output = (x - mean) / std
  assert output.ndim == x.ndim
  if scale is not None:
    assert scale.ndim == 1
    scale = scale.dimshuffle(*(('x',) * (x.ndim - 1) + (0,)))
    assert scale.ndim == x.ndim
    output = output * scale
  if bias is not None:
    assert bias.ndim == 1
    bias = bias.dimshuffle(*(('x',) * (x.ndim - 1) + (0,)))
    assert bias.ndim == x.ndim
    output = output + bias
  return output


def print_to_file(filename, x, argmax=None, sum=None, shape=False):
  def theano_print_to_file(op,x):
    with open(filename, 'a') as f:
      if argmax is not None:
        f.write(str(x.argmax(axis=argmax)) + '\n')
      elif sum is not None:
        f.write(str(x.sum(axis=sum)) + '\n')
      elif shape:
        f.write(str(x.shape) + '\n')
      else:
        #opt = numpy.get_printoptions()
        #numpy.set_printoptions(threshold='nan')
        f.write(str(x) + '\n')
        #numpy.set_printoptions(opt)
  return theano.printing.Print(global_fn=theano_print_to_file)(x)


def self_similarity_cosine(x):
  """
  :param x: shape (T,D)
  :returns cosine similarity matrix, shape (T,T), zeroed at diagonal and upper triangle
  """
  assert x.ndim == 2
  x2 = T.sqrt(T.clip(T.sum(T.sqr(x), axis=1), numpy.float32(1.e-20), numpy.float32(1.e20)))
  assert x2.ndim == 1
  x_ = x / x2.dimshuffle(0, 'x')
  xx = T.dot(x_, x_.T)
  assert xx.ndim == 2
  from theano.tensor.basic import tril
  return tril(xx, 1)



# -*- coding: utf8 -*-

from __future__ import print_function

import _setup_test_env  # noqa
import unittest
from nose.tools import assert_equal, assert_is_instance, assert_in, assert_not_in, assert_true, assert_false
from returnn.datasets.generating import *
from returnn.datasets.basic import DatasetSeq
from returnn.util.basic import PY3, unicode
import os
import unittest

from returnn.util import better_exchook
from returnn.log import log


my_dir = os.path.dirname(os.path.realpath(__file__))


def test_init():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
  assert_equal(dataset.num_inputs, 2)
  assert_equal(dataset.num_outputs, {"classes": (3, 1), "data": (2, 2)})
  assert_equal(dataset.num_seqs, 4)


def test_load_seqs():
  dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
  dataset.init_seq_order(epoch=1)
  dataset.load_seqs(0, 1)
  dataset.load_seqs(1, 3)


@unittest.skipIf(not os.path.exists("/tmp/enwik8.zip"), "we will not trigger the download")
def test_Enwik8Corpus_batch_num_seqs():
  dataset = Enwik8Corpus(path="/tmp", subset="validation", seq_len=13)
  dataset.init_seq_order(epoch=17)
  data = b""
  n = 0
  while dataset.is_less_than_num_seqs(n) and n < 100:
    dataset.load_seqs(n, n + 1)
    data += bytes(dataset.get_data(n, "data"))
    n += 1

  batch_size = 23
  batch_data = [b"" for i in range(batch_size)]
  dataset = Enwik8Corpus(path="/tmp", subset="validation", seq_len=9, batch_num_seqs=batch_size)
  dataset.init_seq_order(epoch=31)
  n = 0
  while dataset.is_less_than_num_seqs(n) and n < 100:
    dataset.load_seqs(n, n + 1)
    new_data = bytes(dataset.get_data(n, "data"))
    batch_data[n % batch_size] += new_data
    n += 1
  assert data.startswith(batch_data[0])


def test_StaticDataset_custom_keys():
  dataset = StaticDataset([{"source": numpy.array([1, 2, 3]), "target": numpy.array([3, 4, 5, 6, 7])}])
  dataset.init_seq_order(epoch=1)
  assert dataset.num_seqs == 1
  assert_equal(dataset.get_data_keys(), ["source", "target"])
  assert_equal(dataset.num_outputs["source"][1], 1)
  assert_equal(dataset.num_outputs["target"][1], 1)
  dataset.load_seqs(0, 1)
  assert_equal(list(dataset.get_data(0, "source")), [1, 2, 3])
  assert_equal(list(dataset.get_data(0, "target")), [3, 4, 5, 6, 7])


def test_StaticDataset_custom_keys_with_dims():
  dataset = StaticDataset(
    data=[{"source": numpy.array([1, 2, 3]), "target": numpy.array([3, 4, 5, 6, 7])}],
    output_dim={"source": [5, 1], "target": [10, 1]})
  dataset.init_seq_order(epoch=1)
  assert dataset.num_seqs == 1
  assert_equal(dataset.get_data_keys(), ["source", "target"])
  assert_equal(dataset.num_outputs["source"][1], 1)
  assert_equal(dataset.num_outputs["target"][1], 1)
  dataset.load_seqs(0, 1)
  assert_equal(list(dataset.get_data(0, "source")), [1, 2, 3])
  assert_equal(list(dataset.get_data(0, "target")), [3, 4, 5, 6, 7])


def test_StaticDataset_utf8():
  s = u"wër"
  print("some unicode str:", s, "repr:", repr(s), "type:", type(s), "len:", len(s))
  assert len(s) == 3
  if PY3:
    assert isinstance(s, str)
    s_byte_list = list(s.encode("utf8"))
  else:
    assert isinstance(s, unicode)
    s_byte_list = list(map(ord, s.encode("utf8")))
  print("utf8 byte list:", s_byte_list)
  assert len(s_byte_list) == 4 > 3
  raw = numpy.array(s_byte_list, dtype="uint8")
  assert_equal(raw.tolist(), [119, 195, 171, 114])
  data = StaticDataset([{"data": raw}], output_dim={"data": (255, 1)})
  if "data" not in data.labels:
    data.labels["data"] = [chr(i) for i in range(255)]  # like in SprintDataset
  data.init_seq_order(epoch=1)
  data.load_seqs(0, 1)
  raw_ = data.get_data(seq_idx=0, key="data")
  assert_equal(raw.tolist(), raw_.tolist())
  assert data.can_serialize_data(key="data")
  s_serialized = data.serialize_data(key="data", data=raw)
  print("serialized:", s_serialized, "repr:", repr(s_serialized), "type:", type(s_serialized))
  assert_equal(s, s_serialized)


# might be moved to a separate test_MetaDataset ...
def test_ConcatSeqsDataset():
  num_seqs = 2
  seq_len = 3
  sub_dataset = StaticDataset(
    [{"data": numpy.array(range(1, seq_len + 1))}] * num_seqs
  )
  from returnn.datasets.meta import ConcatSeqsDataset
  import tempfile
  seq_list_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-list", suffix=".txt")
  seq_len_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-len", suffix=".txt")
  with seq_list_f, seq_len_f:
    seq_len_f.write("{\n")
    for i in range(num_seqs):
      seq_list_f.write("seq-%i\n" % i)
      seq_len_f.write("'seq-%i': %i,\n" % (i, seq_len))
    seq_len_f.write("}\n")
    for i in range(0, num_seqs, 2):
      seq_list_f.write("seq-%i;seq-%i\n" % (i, i + 1))
    seq_list_f.flush()
    seq_len_f.flush()
    dataset = ConcatSeqsDataset(
      dataset=sub_dataset,
      seq_list_file=seq_list_f.name, seq_len_file=seq_len_f.name)
    dataset.init_seq_order(epoch=1)
    concat_num_seqs = num_seqs + num_seqs // 2
    dataset.load_seqs(0, concat_num_seqs)
    assert dataset.num_seqs == concat_num_seqs == 3
    assert_equal(dataset.get_data(0, "data").tolist(), [1, 2, 3])
    assert_equal(dataset.get_data(1, "data").tolist(), [1, 2, 3])
    assert_equal(dataset.get_data(2, "data").tolist(), [1, 2, 3, 1, 2, 3])


# might be moved to a separate test_MetaDataset ...
def test_ConcatSeqsDataset_repeat_in_between_last_frame_up_to_multiple_of():
  num_seqs = 2
  sub_dataset = StaticDataset(
    [{"data": numpy.array([1, 2])}, {"data": numpy.array([1, 2, 3])}]
  )
  from returnn.datasets.meta import ConcatSeqsDataset
  import tempfile
  seq_list_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-list", suffix=".txt")
  seq_len_f = tempfile.NamedTemporaryFile(mode="w", prefix="seq-len", suffix=".txt")
  with seq_list_f, seq_len_f:
    seq_len_f.write("%r\n" % ({"seq-0": 2, "seq-1": 3},))
    seq_len_f.flush()
    seq_list_f.write("seq-0\n")
    seq_list_f.write("seq-1\n")
    seq_list_f.write("seq-0;seq-1;seq-1\n")
    seq_list_f.flush()
    concat_num_seqs = 3
    dataset = ConcatSeqsDataset(
      dataset=sub_dataset,
      repeat_in_between_last_frame_up_to_multiple_of={"data": 5},
      seq_list_file=seq_list_f.name, seq_len_file=seq_len_f.name)
    dataset.init_seq_order(epoch=1)
    dataset.load_seqs(0, concat_num_seqs)
    assert dataset.num_seqs == concat_num_seqs == 3
    assert_equal(dataset.get_data(0, "data").tolist(), [1, 2])
    assert_equal(dataset.get_data(1, "data").tolist(), [1, 2, 3])
    assert_equal(dataset.get_data(2, "data").tolist(), [1, 2, 2, 2, 2, 1, 2, 3, 3, 3, 1, 2, 3])


def test_BytePairEncoding_unicode():
  bpe = BytePairEncoding(
    bpe_file="%s/bpe-unicode-demo.codes" % my_dir,
    vocab_file="%s/bpe-unicode-demo.vocab" % my_dir,
    unknown_label="<unk>")
  assert_equal(bpe.num_labels, 189)
  assert_equal(bpe.labels[5], "z")
  assert_equal(bpe.vocab["z"], 5)
  assert_equal(bpe.bpe._bpe_codes[("n", "d</w>")], 1)
  assert_equal(bpe.labels[6], u"å")
  assert_equal(bpe.vocab[u"å"], 6)
  assert_equal(bpe.bpe._bpe_codes[(u"à", u"nd</w>")], 2)

  def get_bpe_seq(text):
    """
    :param str text:
    :rtype: str
    """
    bpe_label_seq = bpe.get_seq(text)
    res = " ".join(bpe.labels[i] for i in bpe_label_seq)
    print("%r -> %r" % (text, res))
    return res

  assert_equal(get_bpe_seq("kod"), "k@@ o@@ d")  # str
  assert_equal(get_bpe_seq(u"kod"), u"k@@ o@@ d")  # unicode
  assert_equal(get_bpe_seq(u"råt"), u"råt")
  assert_equal(
    get_bpe_seq(u"råt råt iz ďër iz ďër ám àn iz ďër ë låk ë kod áv dres wër yù wêk dù ďë àsk"),
    u"råt råt iz ďër iz ďër ám à@@ n iz ďër ë låk ë k@@ o@@ d áv d@@ r@@ e@@ s w@@ ër yù w@@ ê@@ k dù ďë à@@ s@@ k")


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from __future__ import print_function

import sys
import os

import _setup_test_env  # noqa
import unittest
import theano
import theano.tensor as T
import numpy
from returnn.util.basic import have_gpu
import RecurrentTransform
import theano.sandbox.cuda as cuda
import CustomLSTMFunctions

CustomLSTMFunctions.debug_function_hook = True


def get_attention(att_class, **kwargs):
  import OpLSTMCustom
  recurrent_transform = RecurrentTransform.get_dummy_recurrent_transform(att_class.name, **kwargs)
  assert isinstance(recurrent_transform, att_class)
  f = OpLSTMCustom.register_func(recurrent_transform)
  return f

LSTMCustomTestOpNoInplaceInstance = get_attention(RecurrentTransform.AttentionTest)
LSTMCustomDotAttentionOpNoInplaceInstance = get_attention(RecurrentTransform.AttentionDot)
from OpLSTM import LSTMOpInstance


@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_does_not_crash():
  Z = T.ftensor3('Z')
  W_re = T.fmatrix('W_re')
  W_att_in = T.fmatrix('W_att_in')
  c = T.fmatrix('c') #initial state
  y0 = T.fmatrix('y0') #initial activation
  i = T.matrix('i',dtype='int8')
  Y, H, d = LSTMCustomTestOpNoInplaceInstance(Z, c, y0, i, W_re, W_att_in)

  f = theano.function(inputs=[Z, c, y0, i, W_re, W_att_in], outputs=Y)

  n_T = 5
  n_batch = 4
  n_inp_dim = 3
  n_cells = 8
  numpy.random.seed(1234)
  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  #i_val = numpy.ones((n_T, n_batch), dtype='int8')
  i_val = numpy.array([[1,1,1,1,1], [0,0,1,1,1], [0,0,1,1,1], [0,0,1,0,0]], dtype='int8').T

  Y_val = numpy.asarray(f(Z_val, c_val, y0_val, i_val, W_re_val, W_att_in_val))
  #print Y_val
  print("success")

@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_fwd_pass_compatible_with_OpLSTM():
  Z = T.ftensor3('Z')
  W_re = T.fmatrix('W_re')
  W_att_in = T.fmatrix('W_att_in')
  c = T.fmatrix('c') #initial state
  y0 = T.fmatrix('y0') #initial activation
  i = T.matrix('i',dtype='int8')

  Y, H, d = LSTMCustomTestOpNoInplaceInstance(Z, c, y0, i, W_re, W_att_in)
  W_re_modified = W_re + W_att_in
  Z_modified = T.inc_subtensor(Z[0], T.dot(y0,W_re_modified))
  Y2, H2, d2 = LSTMOpInstance(Z_modified, W_re_modified, c, i)

  f = theano.function(inputs=[Z, c, y0, i, W_re, W_att_in], outputs=Y)
  g = theano.function(inputs=[Z, W_re, c, y0, i, W_att_in], outputs=Y2)

  n_T = 5
  n_batch = 4
  n_inp_dim = 3
  n_cells = 8
  numpy.random.seed(1234)
  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  #i_val = numpy.ones((n_T, n_batch), dtype='int8')
  i_val = numpy.array([[1,1,1,1,1], [0,0,1,1,1], [0,0,1,1,1], [0,0,1,0,0]], dtype='int8').T

  Y_val = numpy.asarray(f(Z_val, c_val, y0_val, i_val, W_re_val, W_att_in_val))
  Y2_val = numpy.asarray(g(Z_val, W_re_val, c_val, y0_val, i_val, W_att_in_val))
  assert numpy.allclose(Y_val, Y2_val)
  print("success")

@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_bwd_pass_compatible_with_OpLSTM():
  Z = T.ftensor3('Z')
  W_re = T.fmatrix('W_re')
  W_att_in = T.fmatrix('W_att_in')
  c = T.fmatrix('c') #initial state
  y0 = T.fmatrix('y0') #initial activation
  i = T.matrix('i',dtype='int8')
  Y, H, d = LSTMCustomTestOpNoInplaceInstance(Z, c, y0, i, W_re, W_att_in)
  W_re_modified = W_re + W_att_in
  Z_modified = T.inc_subtensor(Z[0], T.dot(y0,W_re_modified))
  Y2, H2, d2 = LSTMOpInstance(Z_modified, W_re_modified, c, i)

  cost = Y.sum()
  DZ = T.grad(cost, Z)
  DW_re = T.grad(cost, W_re)
  DW_att_in = T.grad(cost, W_att_in)
  Dc = T.grad(cost, c)
  Dy0 = T.grad(cost, y0)
  cost2 = Y2.sum()
  DZ2 = T.grad(cost2, Z)
  DW_re2 = T.grad(cost2, W_re)
  DW_att_in2 = T.grad(cost2, W_att_in)
  Dc2 = T.grad(cost2, c)
  Dy02 = T.grad(cost2, y0)

  f = theano.function(inputs=[Z, c, y0, i, W_re, W_att_in], outputs=[DZ, DW_re, Dc, Dy0, DW_att_in])
  g = theano.function(inputs=[Z, W_re, c, y0, i, W_att_in], outputs=[DZ2, DW_re2, Dc2, Dy02, DW_att_in2])

  n_T = 5
  n_batch = 4
  n_inp_dim = 3
  n_cells = 8
  numpy.random.seed(1234)
  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  #W_att_in_val = numpy.zeros((n_cells, 4 * n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  #i_val = numpy.ones((n_T, n_batch), dtype='int8')
  i_val = numpy.array([[1,1,1,1,1], [0,0,1,1,1], [0,0,1,1,1], [0,0,1,0,0]], dtype='int8').T

  vals = f(Z_val, c_val, y0_val, i_val, W_re_val, W_att_in_val)
  DZ_val, DW_re_val, Dc_val, Dy0_val, DW_att_in_val = [numpy.asarray(x) for x in vals]
  vals2 = g(Z_val, W_re_val, c_val, y0_val, i_val, W_att_in_val)
  DZ2_val, DW_re2_val, Dc2_val, Dy02_val, DW_att_in2_val = [numpy.asarray(x) for x in vals2]
  assert numpy.allclose(DZ_val, DZ2_val, atol=5e-7, rtol=1e-4)
  assert numpy.allclose(DW_re_val, DW_re2_val, atol=5e-7, rtol=1e-4)
  assert numpy.allclose(Dc_val, Dc2_val)
  assert numpy.allclose(Dy0_val, Dy02_val)
  assert numpy.allclose(DW_att_in_val, DW_att_in2_val, atol=5e-7, rtol=1e-4)
  print("success")

@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_grads():
  n_T = 5
  n_batch = 4
  n_inp_dim = 3
  n_cells = 8
  numpy.random.seed(1234)
  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  #i_val = numpy.ones((n_T, n_batch), dtype='int8')
  i_val = numpy.array([[1,1,1,1,1], [0,0,1,1,1], [0,0,1,1,1], [0,0,1,0,0]], dtype='int8').T

  print("verifying grads...")

  #ignore W_att_in atm
  def LSTMCustomOp_Z_onlyZ(Z):
      return LSTMCustomTestOpNoInplaceInstance(Z, c_val, y0_val, i_val, W_re_val, W_att_in_val)[0]

  def LSTMCustomOp_Z(Z, c, y0, W_re, W_att_in):
    return LSTMCustomTestOpNoInplaceInstance(Z, c, y0, i_val, W_re, W_att_in)[0]

  def LSTMCustomOp_d(Z, c, y0, W_re, W_att_in):
    return LSTMCustomTestOpNoInplaceInstance(Z, c, y0, i_val, W_re, W_att_in)[2]

  print("verifying grad of Z (only w.r.t. Z)")
  theano.tests.unittest_tools.verify_grad(LSTMCustomOp_Z_onlyZ, [Z_val])
  print("verifying grad of Z")
  theano.tests.unittest_tools.verify_grad(LSTMCustomOp_Z, [Z_val, c_val, y0_val, W_re_val, W_att_in_val])
  print("verifying grad of d")
  theano.tests.unittest_tools.verify_grad(LSTMCustomOp_d, [Z_val, c_val, y0_val, W_re_val, W_att_in_val], eps=1e-3)

  print("success")

#---- attention functions

@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_attention_dot_does_not_crash():
  Z = T.ftensor3('Z')
  B = T.ftensor3('B') #base
  W_re = T.fmatrix('W_re')
  W_att_quadr = T.fmatrix("W_att_quadr")
  W_att_in = T.fmatrix('W_att_in')
  c = T.fmatrix('c') #initial state
  y0 = T.fmatrix('y0') #initial activation
  i = T.matrix('i',dtype='int8')
  Y, H, d = LSTMCustomDotAttentionOpNoInplaceInstance(Z, c, y0, i, W_re, B, W_att_in, W_att_quadr)

  f = theano.function(inputs=[Z, B, c, y0, i, W_re, W_att_in, W_att_quadr], outputs=Y)

  n_B = 8
  n_T = 5
  n_batch = 4
  n_cells = 8
  numpy.random.seed(1234)
  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  B_val = numpy.random.ranf((n_B,n_batch,n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_quadr_val = numpy.eye(n_B).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  #i_val = numpy.ones((n_T, n_batch), dtype='int8')
  i_val = numpy.array([[1,1,1,1,1], [0,0,1,1,1], [0,0,1,1,1], [0,0,1,0,0]], dtype='int8').T

  Y_val = numpy.asarray(f(Z_val, B_val, c_val, y0_val, i_val, W_re_val, W_att_in_val, W_att_quadr_val))
  #print Y_val
  print("success")

@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_attention_dot_grads():
  n_T = 5
  n_batch = 4
  n_inp_dim = 3
  n_cells = 8
  n_B = 8
  numpy.random.seed(1234)
  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_quadr_val = numpy.eye(n_B).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  B_val = numpy.random.ranf((n_B,n_batch,n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  #i_val = numpy.ones((n_T, n_batch), dtype='int8')
  i_val = numpy.array([[1,1,1,1,1], [0,0,1,1,1], [0,0,1,1,1], [0,0,1,0,0]], dtype='int8').T

  print("verifying grads...")

  def LSTMCustomOp_Z(Z, c, y0, W_re, B, W_att_in, W_att_quadr):
    return LSTMCustomDotAttentionOpNoInplaceInstance(Z, c, y0, i_val, W_re, B, W_att_in, W_att_quadr)[0]

  def LSTMCustomOp_d(Z, c, y0, W_re, B, W_att_in, W_att_quadr):
    return LSTMCustomDotAttentionOpNoInplaceInstance(Z, c, y0, i_val, W_re, B, W_att_in, W_att_quadr)[2]

  print("verifying grad of Z")
  theano.tests.unittest_tools.verify_grad(LSTMCustomOp_Z, [Z_val, c_val, y0_val, W_re_val, B_val, W_att_in_val, W_att_quadr_val])
  print("verifying grad of d")
  theano.tests.unittest_tools.verify_grad(LSTMCustomOp_d, [Z_val, c_val, y0_val, W_re_val, B_val, W_att_in_val, W_att_quadr_val], eps=1e-3)

  print("success")


@unittest.skipIf(not have_gpu(), "no gpu on this system")
def test_attention_time_gauss():
  n_T = 4
  n_batch = 2
  n_inp_dim = 3
  n_cells = 5
  n_B = 5

  custom_op = get_attention(RecurrentTransform.AttentionTimeGauss,
                            n_out=n_cells, n_batches=n_batch, n_input_t=n_B, n_input_dim=n_inp_dim)
  att = custom_op.recurrent_transform

  Z_val = numpy.random.ranf((n_T,n_batch,4*n_cells)).astype('float32')
  W_re_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  W_att_quadr_val = numpy.eye(n_B).astype('float32')
  W_att_in_val = numpy.random.ranf((n_cells, 4 * n_cells)).astype('float32')
  B_val = numpy.random.ranf((n_B,n_batch,n_cells)).astype('float32')
  c_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  y0_val = numpy.random.ranf((n_batch, n_cells)).astype('float32')
  i_val = numpy.ones((n_T, n_batch), dtype='int8')

  Z = T.ftensor3('Z')
  B = T.ftensor3('B') #base
  W_re = T.fmatrix('W_re')
  W_att_quadr = T.fmatrix("W_att_quadr")
  W_att_in = T.fmatrix('W_att_in')
  c = T.fmatrix('c') #initial state
  y0 = T.fmatrix('y0') #initial activation
  i = T.matrix('i',dtype='int8')
  t0 = T.fvector('t0')
  custom_vars = att.get_sorted_custom_vars()
  initial_state_vars = att.get_sorted_state_vars_initial()
  custom_op_inputs = [Z, c, y0, i, W_re] + custom_vars + initial_state_vars
  print("input args num:", len(custom_op_inputs))
  print("input args:", custom_op_inputs)
  custom_op_outputs = custom_op(*custom_op_inputs)
  print("output args num:", len(custom_op_outputs))
  custom_op_outputs = [cuda.host_from_gpu(v) for v in custom_op_outputs]
  f = theano.function(inputs=[Z, c, y0, i, W_re], outputs=custom_op_outputs)

  res = f(Z_val, c_val, y0_val, i_val, W_re_val)

  #print res
  # res: (output) Y, (gates and cell state) H, (final cell state) d, state vars sequences
  (Y, H, d), state_var_seqs = res[:3], res[3:]

  # print "running custom dumped data"
  # custom_op_inputs = [theano.shared(numpy.load("../op.i.%i" % i)) for i in range(12)]
  # custom_op_outputs = custom_op(*custom_op_inputs)
  # custom_op_outputs = [cuda.host_from_gpu(v) for v in custom_op_outputs]
  # f = theano.function(inputs=[], outputs=custom_op_outputs)
  # res = f()

  print(res)

  assert False


if __name__ == '__main__':
  #tests with attention
  print("calling test_attention_dot_does_not_crash()")
  test_attention_dot_does_not_crash()
  print("calling test_attention_dot_grads")
  test_attention_dot_grads()

  #tests without attention
  print("calling test_does_not_crash()")
  test_does_not_crash()
  print("calling test_fwd_pass_compatible_with_OpLSTM()")
  test_fwd_pass_compatible_with_OpLSTM()
  print("calling test_bwd_pass_compatible_with_OpLSTM()")
  test_bwd_pass_compatible_with_OpLSTM()
  print("calling test_grads()")
  test_grads()



"""
Provides :class:`MapDatasetBase`
"""

from returnn.datasets.basic import DatasetSeq
from returnn.datasets.cached2 import CachedDataset2
from returnn.util.basic import OptionalNotImplementedError


class MapDatasetBase(object):
  """
  This dataset can be used as template to implement user-side Datasets, where the data can be access in arbitrary order.
  For global sorting, the length information needs to be known beforehand, see get_seq_len.
  """

  def __init__(self, data_types=None):
    """
    :param dict[str,dict] data_types: data_key -> constructor parameters of Data object, for all data streams the
      dataset provides (inputs and targets). E.g. {'data': {'dim': 1000, 'sparse': True, ...}, 'classes': ...}.
    """
    self.data_types = data_types or {}

  def __len__(self):
    """
    :return: total number of sequences in the dataset
    :rtype: int
    """
    raise NotImplementedError

  def __getitem__(self, seq_idx):
    """
    This function does the actual data loading, the order can be arbitrary.

    :param int seq_idx:
    :return: The content of a single dataset entry
    :rtype: dict[str,numpy.array]
    """
    raise NotImplementedError

  def get_seq_len(self, seq_idx):
    """
    This optional function provides the sequence length for the `seq_ordering` parameter.
    If not specified only a limited set of options is available.

    :param int seq_idx:
    :return: sequence length
    :rtype: int
    """
    raise OptionalNotImplementedError

  def get_seq_tag(self, seq_idx):
    """
    :param int seq_idx:
    :return: tag for the sequence of the given index, default is 'seq-{seq_idx}'.
    :rtype: str
    """
    return "seq-%i" % seq_idx

  def get_seq_order(self, epoch=None):
    """
    Override to implement a dataset specific sequence order for a given epoch.
    The number of sequences can be less than the total number.
    This will override the effects of `partition_epoch` and `seq_ordering` when using MapDatasetWrapper.

    :param int epoch:
    :return: sequence order (list of sequence indices)
    :rtype: list[int]
    """
    raise OptionalNotImplementedError


class MapDatasetWrapper(CachedDataset2):
  """
  Takes a MapDataset and turns it into a returnn.datasets.Dataset by providing the required class methods.
  """
  def __init__(self, map_dataset, **kwargs):
    """
    :param MapDatasetBase|function map_dataset: the MapDataset to be wrapped
    """
    super(MapDatasetWrapper, self).__init__(**kwargs)

    if callable(map_dataset):
      map_dataset = map_dataset(**kwargs)
    assert isinstance(map_dataset, MapDatasetBase)

    self._dataset = map_dataset
    self._seq_order = None

  @property
  def num_seqs(self):
    """
    :returns number of sequences in the current epoch
    :rtype: int
    """
    if self._seq_order is None:
      raise NotImplementedError("'num_seqs' is only known after calling init_seq_order().")
    return len(self._seq_order)

  def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
    """
    :param int|None epoch:
    :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
    :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order.
    :rtype: bool
    :returns whether the order changed (True is always safe to return)
    """
    super(MapDatasetWrapper, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

    if seq_list or seq_order:
      raise NotImplementedError

    try:
      self._seq_order = self._dataset.get_seq_order(epoch=epoch)
    except OptionalNotImplementedError:
      try:
        self._seq_order = self.get_seq_order_for_epoch(
          epoch=epoch, num_seqs=len(self._dataset), get_seq_len=self._dataset.get_seq_len)
      except OptionalNotImplementedError:
        # only support seq_ordering that need no length here
        assert self.seq_ordering in ["default", "reverse", "random"]
        self._seq_order = self.get_seq_order_for_epoch(
          epoch=epoch, num_seqs=len(self._dataset), get_seq_len=None)

    return True

  def _collect_single_seq(self, seq_idx):
    """
    :param int seq_idx: sorted seq idx
    :return:
    """
    corpus_seq_idx = self.get_corpus_seq_idx(seq_idx)
    return DatasetSeq(
      seq_idx, features=self._dataset[corpus_seq_idx], seq_tag=self._dataset.get_seq_tag(corpus_seq_idx))

  def get_current_seq_order(self):
    """
    :rtype: typing.Sequence[int]
    """
    assert self._seq_order is not None
    return self._seq_order

  def get_tag(self, sorted_seq_idx):
    """
    :param sorted_seq_idx:
    :return:
    """
    seq_tag = self._dataset.get_seq_tag(self.get_corpus_seq_idx(sorted_seq_idx))
    return seq_tag

  def get_corpus_seq_idx(self, sorted_seq_idx):
    """
    :param int sorted_seq_idx:
    :return corpus_seq_idx
    :rtype: int
    """
    return self._seq_order[sorted_seq_idx]

  def have_corpus_seq_idx(self):
    """
    :rtype: bool
    :return: whether you can call self.get_corpus_seq_idx()
    """
    return True

  def get_data_dim(self, key):
    """
    :param str key: e.g. "data" or "classes"
    :return: number of classes, no matter if sparse or not
    :rtype: int
    """
    if key in self._dataset.data_types:
      return self._dataset.data_types[key].get("dim", 1)
    return 1  # unknown

  def get_data_dtype(self, key):
    """
    :param str key: e.g. "data" or "classes"
    :return: dtype as str, e.g. "int32" or "float32"
    :rtype: str
    """
    if key in self._dataset.data_types and "dtype" in self._dataset.data_types[key]:
      return self._dataset.data_types[key]["dtype"]
    if self.is_data_sparse(key):
      return "int32"
    return "float32"

  def is_data_sparse(self, key):
    """
    :param str key: e.g. "data" or "classes"
    :return: whether the data is sparse
    :rtype: bool
    """
    # Note: We cannot call get_data_dtype, as we would maybe result in infinite recursion.
    if key in self._dataset.data_types:
      return self._dataset.data_types[key].get("sparse", False)
    return False

  def get_data_shape(self, key):
    """
    :returns get_data(*, key).shape[1:], i.e. num-frames excluded
    :rtype: list[int]
    """
    if key in self._dataset.data_types:
      if "shape" in self._dataset.data_types[key].keys():
        if self._dataset.data_types[key]["shape"][0] is None:
          return self._dataset.data_types[key]["shape"][1:]
        else:
          assert False, "data shape has no time axis, calling get_data_shape is not possible"
    if self.is_data_sparse(key):
      return []
    return [self.get_data_dim(key)]


class FromListDataset(MapDatasetBase):
  """
  Simple implementation of a MapDataset where all data is given in a list.
  """
  def __init__(self, data_list, sort_data_key=None, **kwargs):
    """
    :param list[dict[str,numpy.ndarray]] data_list: sequence data as a dict data_key -> data for all sequences.
    :param str sort_data_key: Sequence length will be determined from data of this data_key.
    """
    self._data_list = data_list
    self._sort_data_key = sort_data_key
    super(FromListDataset, self).__init__(**kwargs)

  def __len__(self):
    """
    :return: total number of sequences in the dataset
    :rtype: int
    """
    return len(self._data_list)

  def __getitem__(self, seq_idx):
    """
    :param int seq_idx:
    :return: The content of a single dataset entry
    :rtype dict[str,numpy.array]
    """
    return self._data_list[seq_idx]

  def get_seq_len(self, seq_idx):
    """
    :param seq_idx:
    :return: length of data for 'sort_data_key'
    :rtype: int
    """
    assert self._sort_data_key, "Specify which data key should be used for sequence sorting via 'sort_data_key'."
    return len(self._data_list[seq_idx][self._sort_data_key])
